﻿using OlaCudafy1;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Cudafy;
using OlaCudafy1.FDTD;
using OlaCudafy1.Material;
using Cudafy.Host;
using System.Drawing;

namespace OlaCudafy1.Test
{
    
    
    /// <summary>
    ///This is a test class for FDTDDrawMachineTest and is intended
    ///to contain all FDTDDrawMachineTest Unit Tests
    ///</summary>
	[TestClass()]
	public class FDTDDrawMachineTest
	{


		private TestContext testContextInstance;

		/// <summary>
		///Gets or sets the test context which provides
		///information about and functionality for the current test run.
		///</summary>
		public TestContext TestContext
		{
			get
			{
				return testContextInstance;
			}
			set
			{
				testContextInstance = value;
			}
		}

		#region Additional test attributes
		// 
		//You can use the following additional attributes as you write your tests:
		//
		//Use ClassInitialize to run code before running the first test in the class
		//[ClassInitialize()]
		//public static void MyClassInitialize(TestContext testContext)
		//{
		//}
		//
		//Use ClassCleanup to run code after all tests in a class have run
		//[ClassCleanup()]
		//public static void MyClassCleanup()
		//{
		//}
		//
		//Use TestInitialize to run code before running each test
		//[TestInitialize()]
		//public void MyTestInitialize()
		//{
		//}
		//
		//Use TestCleanup to run code after each test has run
		//[TestCleanup()]
		//public void MyTestCleanup()
		//{
		//}
		//
		#endregion


		/// <summary>
		///A test for SetHSV
		///</summary>
		[TestMethod()]
		public void SetHSVTest()
		{
			// testing with example color #A0A424,
			//   data taken from http://en.wikipedia.org/wiki/HSL_and_HSV#Examples
			// 0.628	0.643	0.142	61.8°    0.779	0.643	
			// r         g        b     h (deg)  s_hsv    v
			
			// converting
			const float h = 61.8F/360F;
			const float s = 0.779F;
			const float v = 0.643F;

			// should convert to
			const float rF = 0.628F;
			const byte rExpected = (byte) (rF*256);
			const float gF = 0.643F;
			const byte gExpected = (byte) (gF*256);
			const float bF = 0.142F;
			const byte bExpected = (byte) (bF*256);


			byte rActual;
			byte gActual;
			byte bActual;
			FDTDDrawMachine.SetHSV(h, s, v, out rActual, out gActual, out bActual);

			Assert.AreEqual(rExpected, rActual);
			Assert.AreEqual(gExpected, gActual);
			Assert.AreEqual(bExpected, bActual);
		}

		/// <summary>
		///A test for FDTDDrawKernel. Tests the correct calculation of ONE single pixel.
		///</summary>
		[TestMethod()]
		public void FDTDDrawKernelTest()
		{
			var extent = new Extent(2, 3, 4);
			var p = new float[extent.X, extent.Y, extent.Z];

			var horiz = new float[extent.X, extent.Y, extent.Z];
			var vert = new float[extent.X, extent.Y, extent.Z];
			var zData = new float[extent.X, extent.Y, extent.Z];
			
			const int displaySlice = 1;
			const int sliceDimension = 0; // x i.e. picture's w==y and h==z
			var dim = new dim3(3, 4);
			var gGrid = new GGrid(dim);

			// which thread to simulate
			var threadToSimulateW = 2;
			var threadToSimulateH = 3;
			var thread = new GThread(
				threadToSimulateW, threadToSimulateH, new GBlock(gGrid, dim, 0, 0));
			p[displaySlice, thread.threadIdx.x, thread.threadIdx.y] = 0.020f; // 20 mPa aka 60dB

			const int four = 4; // number of bytes for r, g, b and alpha
			var dest = new byte[3, 4, four]; 
			var mat = new byte[2, 3, 4]; 
			
			// currently not used in drawing
			var srcs = new[] { new Source(0, 0, 0, 440F, 0.020f), }; // 60dB
			var mics = new[] { new Microphone(1, 1, 1), };
			
			var imgEx = (new Extent(3, 4, 1)).KindE;

			var displaySettings = new DisplaySettings(extent)
			                      	{
			                      		WeightPressureInDecibels = 0F,
										WeightBothInDecibels = -28F+60F
			                      	};
			// i.e. p should look like 0.8Pa after scaling
			// (val becomes 0.4 because particle speeds are 0)
			byte unused;
			byte expectedAt2_3_Green;
			const float h = 0.0553333333333F;
			const float s = .83F;
			const float v = .4F;
			FDTDDrawMachine.SetHSV(
				h,
				s,
				v,
				out unused,
				out expectedAt2_3_Green,
				out unused);
			

			FDTDDrawMachine.FDTDDrawKernel(thread,
				p,
				horiz,
				vert,
				zData,
				dest,
				mat,
				displaySettings.ScalePressure,
				displaySettings.ScaleBoth,
				srcs,
				Microphone.ProduceInnerMicrophonesArray(mics),
				displaySlice,
				sliceDimension,
				imgEx);
			int indexOfGreen = 1;
			byte actualAt2_3_Green = dest[2, 3, indexOfGreen];

			Assert.AreEqual(expectedAt2_3_Green, actualAt2_3_Green);
		}

		/// <summary>
		///Tests that Clip clips/clamps values over 1 to 1
		///</summary>
		[TestMethod()]
		public void ClipTest_ClipsOver1()
		{
			float v = 1.2F;
			const float vExpected = 1.0F; // TODO: Initialize to an appropriate value

			FDTDDrawMachine.Clip(ref v);
			
			Assert.AreEqual(vExpected, v);
		}
		/// <summary>
		///Tests that Clip clips/clamps values under 0 to 0
		///</summary>
		[TestMethod()]
		public void ClipTest_ClipsUnder0()
		{
			float v = -0.2F;
			const float vExpected = 0.0F; // TODO: Initialize to an appropriate value

			FDTDDrawMachine.Clip(ref v);

			Assert.AreEqual(vExpected, v);
		}

		/// <summary>
		///A test for ComputeBitmap color mapping
		///</summary>
		[TestMethod()]
		[DeploymentItem("OlaCudafy1.exe")]
		public void ComputeBitmapTest_CorrectColorMapping()
		{
			const int width = 2;
			const int height = 2;
			var fromImage = new byte[width, height, 4];

			byte r = 127;
			byte g = 0;
			byte b = 63;
			byte a = 255;
			fromImage[0, 0, 0] = r;
			fromImage[0, 0, 1] = g;
			fromImage[0, 0, 2] = b;
			fromImage[0, 0, 3] = a;

			Bitmap actual = FDTDDrawMachine_Accessor.ComputeBitmap(fromImage, width, height);

			byte actualColorA = actual.GetPixel(0, 0).A;
			byte actualColorR = actual.GetPixel(0, 0).R;
			byte actualColorG = actual.GetPixel(0, 0).G;
			byte actualColorB = actual.GetPixel(0, 0).B;

			Assert.AreEqual(a, actualColorA);
			Assert.AreEqual(r, actualColorR);
			Assert.AreEqual(g, actualColorG);
			Assert.AreEqual(b, actualColorB);
		}

		/// <summary>
		///A test for ComputeBitmap pixel position mapping
		///</summary>
		[TestMethod()]
		[DeploymentItem("OlaCudafy1.exe")]
		public void ComputeBitmapTest_PixelPositionMapping()
		{
			const int width = 2;
			const int height = 2;
			var fromImage = new byte[width, height, 4];

			Color a = Color.Azure;
			Color b = Color.Blue;
			Color c = Color.Cyan;
			Color d = Color.DarkGray;

			fromImage[0, 0, 0] = a.R;
			fromImage[0, 0, 1] = a.G;
			fromImage[0, 0, 2] = a.B;
			fromImage[0, 0, 3] = 1;

			fromImage[0, 1, 0] = b.R;
			fromImage[0, 1, 1] = b.G;
			fromImage[0, 1, 2] = b.B;
			fromImage[0, 1, 3] = 2;

			fromImage[1, 0, 0] = c.R;
			fromImage[1, 0, 1] = c.G;
			fromImage[1, 0, 2] = c.B;
			fromImage[1, 0, 3] = 3;

			fromImage[1, 1, 0] = d.R;
			fromImage[1, 1, 1] = d.G;
			fromImage[1, 1, 2] = d.B;
			fromImage[1, 1, 3] = 4;
			
			Bitmap actual = FDTDDrawMachine_Accessor.ComputeBitmap(fromImage, width, height);

			Assert.AreEqual(a.R, actual.GetPixel(0, 0).R);
			Assert.AreEqual(a.G, actual.GetPixel(0, 0).G);
			Assert.AreEqual(a.B, actual.GetPixel(0, 0).B);
			Assert.AreEqual(1, actual.GetPixel(0, 0).A);

			Assert.AreEqual(b.R, actual.GetPixel(0, 1).R);
			Assert.AreEqual(b.G, actual.GetPixel(0, 1).G);
			Assert.AreEqual(b.B, actual.GetPixel(0, 1).B);
			Assert.AreEqual(2, actual.GetPixel(0, 1).A);

			Assert.AreEqual(c.R, actual.GetPixel(1, 0).R);
			Assert.AreEqual(c.G, actual.GetPixel(1, 0).G);
			Assert.AreEqual(c.B, actual.GetPixel(1, 0).B);
			Assert.AreEqual(3, actual.GetPixel(1, 0).A);

			Assert.AreEqual(d.R, actual.GetPixel(1, 1).R);
			Assert.AreEqual(d.G, actual.GetPixel(1, 1).G);
			Assert.AreEqual(d.B, actual.GetPixel(1, 1).B);
			Assert.AreEqual(4, actual.GetPixel(1, 1).A);
		}
	}
}
