﻿using OlaCudafy1;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;

namespace OlaCudafy1.Test
{
    
    
    /// <summary>
    ///This is a test class for RindelImpedanceEstimationTest and is intended
    ///to contain all RindelImpedanceEstimationTest Unit Tests
    ///</summary>
	[TestClass()]
	public class RindelImpedanceEstimationTest
	{


		private TestContext testContextInstance;

		/// <summary>
		///Gets or sets the test context which provides
		///information about and functionality for the current test run.
		///</summary>
		public TestContext TestContext
		{
			get
			{
				return testContextInstance;
			}
			set
			{
				testContextInstance = value;
			}
		}

		#region Additional test attributes
		// 
		//You can use the following additional attributes as you write your tests:
		//
		//Use ClassInitialize to run code before running the first test in the class
		//[ClassInitialize()]
		//public static void MyClassInitialize(TestContext testContext)
		//{
		//}
		//
		//Use ClassCleanup to run code after all tests in a class have run
		//[ClassCleanup()]
		//public static void MyClassCleanup()
		//{
		//}
		//
		//Use TestInitialize to run code before running each test
		//[TestInitialize()]
		//public void MyTestInitialize()
		//{
		//}
		//
		//Use TestCleanup to run code after each test has run
		//[TestCleanup()]
		//public void MyTestCleanup()
		//{
		//}
		//
		#endregion


		/// <summary>
		///A test for EstimateResistanceTerm, Testing a Membrane Absorber case
		///</summary>
		[TestMethod()]
		public void EstimateResistanceTermTest_ForMembraneAbsorber()
		{
			var absorptionCoefficients = new[] { 0f, 0.5f, 0.3f, 0.1f, 0.05f, 0.05f };
												// values taken from Figure 5 under section 5.2
												// in Rindels article
			float expected = 2513F; // from Rindel's article

			float acousticImpedanceAt25DegC = 409.4f; // taken from http://en.wikipedia.org/wiki/Density_of_air
			float acousticImpedanceAt5DegC = 424.3f;

			// Calculating to results based on
			// the different values for acoustic impedance one likely
			// could choose.
			float actualAt25DegC =
				(new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedanceAt25DegC)).EstimatedResistanceTerm;
			float actualAt5DegC =
				(new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedanceAt5DegC)).EstimatedResistanceTerm;

			// Checking that 'expected' is between these two results
			float difference = Math.Abs(actualAt25DegC - actualAt5DegC);
			float mean = (actualAt25DegC + actualAt5DegC) / 2f;
			Assert.AreEqual(expected, mean, delta: difference / 2f);
		}

		/// <summary>
		///A test for EstimateResistanceTerm, Testing a Porous Absorber case
		///</summary>
		[TestMethod()]
		public void EstimateResistanceTermTest_ForPorousAbsorber()
		{
			var absorptionCoefficients = new[] { 0f, 0.102f, 0.24f, 0.58f, 0.66f, 0.7f };
												// values taken from Figure 5 under section 5.1
												// in Rindels article
			float expected = 199f; // from Rindel's article

			float acousticImpedanceAt25DegC = 409.4f; // taken from http://en.wikipedia.org/wiki/Density_of_air
			float acousticImpedanceAt5DegC = 424.3f;

			// Calculating to results based on
			// the different values for acoustic impedance one likely
			// could choose.
			float actualAt25DegC =
				(new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedanceAt25DegC)).EstimatedResistanceTerm;
			float actualAt5DegC =
				(new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedanceAt5DegC)).EstimatedResistanceTerm;

			// Checking that 'expected' is between these two results
			float difference = Math.Abs(actualAt25DegC - actualAt5DegC);
			float mean = (actualAt25DegC + actualAt5DegC) / 2f;
			float delta = difference/2f + 20f; // it does not work otherwise. how did Rindel calculate his value for R?
			Assert.AreEqual(expected, mean, delta: delta);
		}


		/// <summary>
		///A test for EstimateResistanceTerm, Testing a Resonating Absorber case
		///</summary>
		[TestMethod()]
		public void EstimateResistanceTermTest_ForResonatorAbsorber()
		{
			var absorptionCoefficients = new[] { 0f, 0.7f, 0.86f, 0.77f, 0.54f, 0.22f }; // from Figure 8
			// values taken from Figure 5 under section 5.1
			// in Rindels article
			float expected = 258f; // from Rindel's article

			float acousticImpedanceAt25DegC = 409.4f; // taken from http://en.wikipedia.org/wiki/Density_of_air
			float acousticImpedanceAt5DegC = 424.3f;

			// Calculating to results based on
			// the different values for acoustic impedance one likely
			// could choose.
			float actualAt25DegC =
				(new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedanceAt25DegC)).EstimatedResistanceTerm;
			float actualAt5DegC =
				(new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedanceAt5DegC)).EstimatedResistanceTerm;

			// Checking that 'expected' is between these two results
			float difference = Math.Abs(actualAt25DegC - actualAt5DegC);
			float mean = (actualAt25DegC + actualAt5DegC) / 2f;
			float delta = difference / 2f + 20f; // it does not work otherwise. how did Rindel calculate his value for R?
			Assert.AreEqual(expected, mean, delta: delta);
		}


		/// <summary>
		///A test for EstimateD; See Rindel's article, latter part of section 5.1
		///</summary>
		[TestMethod()]
		public void EstimateDTest_Porous()
		{
			float[] absorptionCoefficients = new[] { 0f, 0.102f, 0.24f, 0.58f, 0.66f, 0.7f };
			const float acousticImpedance = 413.3F; // 20 deg C
			const float expected = 0.094f;
			const float speedOfSound = 340f;
			float actual = (new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedance)).EstimateD(speedOfSound);
			Assert.AreEqual(expected, actual, delta: 0.005);
		}


		/// <summary>
		///A test for EstimateD for membrane
		///</summary>
		[TestMethod()]
		public void EstimateDTest_Membrane()
		{
			var absorptionCoefficients = new[] { 0f, 0.5f, 0.3f, 0.1f, 0.05f, 0.05f };
			const float acousticImpedance = 413.3F; // 20 deg C
			RindelImpedanceEstimation target =
				new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedance);
			float speedOfSound = 340f;
			float expected = float.PositiveInfinity;
			float actual = target.EstimateD(speedOfSound);
			Assert.AreEqual(expected, actual);
		}

		/// <summary>
		///A test for EstimateM for a membrane absorber
		///</summary>
		[TestMethod()]
		public void EstimateMTest_Membrane()
		{
			var absorptionCoefficients = new[] { 0f, 0.5f, 0.3f, 0.1f, 0.05f, 0.05f };
			const float acousticImpedance = 413.3F; // 20 deg C
			var target =
				new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedance);
			const float expected = 2.139f;
			float actual = target.EstimateM();
			Assert.AreEqual(expected, actual, delta: expected * 0.01); // delta 1% error ok
		}

		/// <summary>
		///A test for EstimateM for a porous absorber
		///</summary>
		[TestMethod()]
		public void EstimateMTest_Porous()
		{
			var absorptionCoefficients = new[] { 0f, 0.102f, 0.24f, 0.58f, 0.66f, 0.7f };
			const float acousticImpedance = 413.3F; // 20 deg C
			var target =
				new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedance);
			const float expected = 0f;
			float actual = target.EstimateM();
			Assert.AreEqual(expected, actual, delta: expected * 0.01); // delta 1% error ok
		}

		/// <summary>
		///A test for EstimateM for Resonator type absorber
		///</summary>
		[TestMethod()]
		public void EstimateMTest_ResonatorM()
		{
			var absorptionCoefficients = new[] { 0f, 0.7f, 0.86f, 0.77f, 0.54f, 0.22f }; // from Figure 8
			const float speedOfSound = 344f;
			const float acousticImpedance = 1.205f * speedOfSound; // 20 deg C
			var target =
				new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedance);
			const float expected = 0.112f;
			float actual = target.EstimateM();
			Assert.AreEqual(expected, actual, delta: expected * 0.01); // delta 1% error ok
		}


		/// <summary>
		///A test for EstimateD for Resonator type absorber
		///</summary>
		[TestMethod()]
		public void EstimateDTest_ResonatorD()
		{
			var absorptionCoefficients = new[] { 0f, 0.7f, 0.86f, 0.77f, 0.54f, 0.22f }; // from Figure 8
			const float speedOfSound = 344f;
			const float acousticImpedance = 1.205f * speedOfSound; // 20 deg C
			var target =
				new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedance);
			const float expected = 0.372f;
			float actual = target.EstimateD(speedOfSound);
			Assert.AreEqual(expected, actual, delta: expected * 0.01); // delta 1% error ok
		}

		/// <summary>
		/// Another test for the R, m and d values of a resonator type absorber
		///</summary>
		[TestMethod()]
		public void EstimateRmd()
		{
			var absorptionCoefficients = new[] { 0f, 0.12f, 0.5f, 0.68f, 0.4f, 0.26f  }; // from Figure 8
			const float speedOfSound = 344f;
			const float acousticImpedance = 1.205f * speedOfSound; // 20 deg C
			var target =
				new RindelImpedanceEstimation(absorptionCoefficients, acousticImpedance);
			const float expectedR = 176f;
			const float expectedM = 0.14f;
			const float expectedD = 0.115f;
			const RindelImpedanceEstimation.AbsorberType expectedAbsorberType
				= RindelImpedanceEstimation.AbsorberType.Resonator;
			float actualR = target.EstimatedResistanceTerm;
			float actualM = target.EstimateM();
			float actualD = target.EstimateD(speedOfSound);
			var actualAbsorberType = RindelImpedanceEstimation.GuessAbsorberType(absorptionCoefficients);
			Assert.AreEqual(expectedR, actualR, delta: expectedR * 0.01); // delta 1% error ok
			Assert.AreEqual(expectedM, actualM, delta: expectedM * 0.01); // delta 1% error ok
			Assert.AreEqual(expectedD, actualD, delta: expectedD * 0.01); // delta 1% error ok
			Assert.AreEqual(expectedAbsorberType, actualAbsorberType);
		}
	}
}
