﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;

namespace OlaCudafy1
{
	internal class AbsorberDatabase
	{
		private const int NumberOfCoefficients = 8;

		public AbsorberDatabase(string path)
		{
			Database = new List<AbsorberDatabaseEntry>();

			// Read in every line in the file.
			using (var reader = new StreamReader(path))
			{
				string line;
				var readState = new ReadState();
				const string headerPattern = @"^\s*([0-9]+)\s(.*)\s*$";
				const string separatorPattern = @"^\s*$";

				int materialNumber = 0;
				string materialDescription = null;

				while ((line = reader.ReadLine()) != null)
				{
					Match match;
					switch (readState.WhichLine)
					{
						case EWhichLine.Default:
							match = Regex.Match(line, headerPattern, RegexOptions.IgnoreCase);

							if (match.Success)
							{
								if (!int.TryParse(match.Groups[1].Value, out materialNumber))
									readState.EntryFailed_NeedsReset = true;
								materialDescription = match.Groups[2].Value;

								readState.Increment();
							}
							else
								readState.Jammed();
							break;

						case EWhichLine.DescriptionLineRead: // read data:
							var coeffs = new float[NumberOfCoefficients];
							bool allParseOK = true;
							int i = 0;
							foreach (string s in line.Split(new[] {'\t', ' '}, 
								StringSplitOptions.RemoveEmptyEntries))
							{
								float result;
								if (!float.TryParse(s.Trim(), out result))
								{
									allParseOK = false;
									break;
								}
								coeffs[i] = result;
								++i;
								if (i >= coeffs.Length)
									break;
							}
							if (allParseOK && !readState.EntryFailed_NeedsReset)
								Database.Add(new AbsorberDatabaseEntry(
								             	materialNumber, materialDescription, coeffs));
							else
								// could not correctly interpret this material specification entry found in file
								readState.Jammed();
							readState.Increment();
							break;
						default: //case EWhichLine.DataLineRead:
							match = Regex.Match(line, separatorPattern);
							if (match.Success)
								readState.Reset();
							break;
					}
				}
			}
		}

		internal List<AbsorberDatabaseEntry> Database { get; set; }

		#region Nested type: EWhichLine

		private enum EWhichLine
		{
			Default,
			DescriptionLineRead,
			WaitingForBlankLine
		}

		#endregion

		#region Nested type: ReadState

		private class ReadState
		{
			public ReadState()
			{
				Reset();
			}

			internal bool EntryFailed_NeedsReset { get; set; }
			internal EWhichLine WhichLine { get; private set; }

			public void Jammed()
			{
				WhichLine = EWhichLine.WaitingForBlankLine;
			}

			internal void Reset()
			{
				EntryFailed_NeedsReset = false;
				WhichLine = EWhichLine.Default;
			}

			internal void Increment()
			{
				if (WhichLine == EWhichLine.Default)
					WhichLine = EWhichLine.DescriptionLineRead;
				else if (WhichLine == EWhichLine.DescriptionLineRead)
					WhichLine = EWhichLine.WaitingForBlankLine;
			}
		}

		#endregion
	}
}