using System;
using System.ComponentModel;

namespace OlaCudafy1
{
	internal class BandCoefficients
	{
		public float Alpha1000HzBand;
		public float Alpha125HzBand;
		public float Alpha250HzBand;
		public float Alpha500HzBand;

		/// <summary>
		/// Sets band coefficients from absorption coefficient array by 
		/// picking out the absorption coefficients to be used in further calculations,
		/// and ignoring the rest.
		///
		/// The supplied float array should be of length 5 but may be longer,
		/// and will be interpreted as follows:
		///	  1st element: 63 Hz: ignored;
		///   2nd-5th element: 125, 250, 500 and 1000 Hz: used for estimation;
		///   any susequent elements if present: ignored;
		/// </summary>
		/// <param name="absorptionCoefficients"></param>
		/// <returns></returns>
		public BandCoefficients(float[] absorptionCoefficients)
		{
			if (absorptionCoefficients == null)
				throw new ArgumentNullException("A non-null argument must be supplied! " +
				                                "Please read the function documentation.");
			if (absorptionCoefficients.Length < 5)
				throw new ArgumentException("The absorption coefficients argument must be of " +
				                            "length 5 or more! " +
				                            "Please read the function documentation.");
			Alpha125HzBand = absorptionCoefficients[1];
			Alpha250HzBand = absorptionCoefficients[2];
			Alpha500HzBand = absorptionCoefficients[3];
			Alpha1000HzBand = absorptionCoefficients[4];
		}

		public BandCoefficients()
		{
		}

		public float Maximum
		{
			get
			{
				return Math.Max(
					Math.Max(
						Math.Max(
							Alpha125HzBand, Alpha250HzBand),
						Alpha500HzBand), Alpha1000HzBand);
			}
		}


		internal float this[FrequencyBand band]
		{
			get
			{
				switch (band)
				{
					case FrequencyBand.Band125Hz:
						return Alpha125HzBand;
					case FrequencyBand.Band250Hz:
						return Alpha250HzBand;
					case FrequencyBand.Band500Hz:
						return Alpha500HzBand;
					case FrequencyBand.Band1000Hz:
						return Alpha1000HzBand;
					default:
						throw new InvalidEnumArgumentException();
				}
			}
			set
			{
				switch (band)
				{
					case FrequencyBand.Band125Hz:
						Alpha125HzBand = value;
						break;
					case FrequencyBand.Band250Hz:
						Alpha250HzBand = value;
						break;
					case FrequencyBand.Band500Hz:
						Alpha500HzBand = value;
						break;
					case FrequencyBand.Band1000Hz:
						Alpha1000HzBand = value;
						break;
					default:
						throw new InvalidEnumArgumentException();
				}
			}
		}

		internal FrequencyBand BandOfMaximum
		{
			get
			{
				if (Alpha125HzBand >= Alpha250HzBand &&
				    Alpha125HzBand >= Alpha500HzBand &&
				    Alpha125HzBand >= Alpha1000HzBand)
					return FrequencyBand.Band125Hz;
				if (Alpha250HzBand >= Alpha125HzBand &&
				    Alpha250HzBand >= Alpha500HzBand &&
				    Alpha250HzBand >= Alpha1000HzBand)
					return FrequencyBand.Band250Hz;
				if (Alpha500HzBand >= Alpha125HzBand &&
				    Alpha500HzBand >= Alpha250HzBand &&
				    Alpha500HzBand >= Alpha1000HzBand)
					return FrequencyBand.Band500Hz;
				if (Alpha1000HzBand >= Alpha125HzBand &&
				    Alpha1000HzBand >= Alpha250HzBand &&
				    Alpha1000HzBand >= Alpha500HzBand)
					return FrequencyBand.Band1000Hz;
				throw new NotImplementedException("Something went wrong");
			}
		}

		internal float[] ToArray()
		{
			var X = new float[5];
			X[0] = 0f;
			X[1] = Alpha125HzBand;
			X[2] = Alpha250HzBand;
			X[3] = Alpha500HzBand;
			X[4] = Alpha1000HzBand;
			return X;
		}
	}
}