﻿using System;
using System.Data;
using Cudafy;
using Cudafy.Host;
using OlaCudafy1.Material;

namespace OlaCudafy1.FDTD
{
	internal class CSimulationSpace : IDisposable
	{
		internal readonly dim3 _blocks;
		private readonly GPGPU _gpu;
		internal readonly dim3 _threads;
		internal KindExtent KindE;
		internal CSoundField TheSoundField;

		/// <summary>
		/// 0 when no wall
		/// 1-255 when wall with absorption number 1-255
		/// </summary>
		internal Wall<byte> TheWalls;

		private GridSettings[] _theGridSettingsArray;
		private GridSettings[] _devTheGridSettingsArray;

		private Source[] _theSources;
		private Source[] _devTheSources;

		private Microphone[] _theMicrophones;
		private MicrophoneStruct[] _devTheMicrophoneStructs;

		public GridSettings GridSettings
		{
			get
			{
				if (_theGridSettingsArray == null || _theGridSettingsArray.Length < 1)
					throw new Exception("GridSettings was not set, for some reason. Review constructor!");
				return _theGridSettingsArray[0];
			}
		}
		public CSimulationSpace(KindExtent kindExtent, GPGPU gpu, Source[] theSources, Microphone[] microphones,
		                        GridSettings gridSettings)
		{
			if (kindExtent.X == 0 &&
			    kindExtent.Y == 0 &&
			    kindExtent.Z == 0)
				throw new ConstraintException("Extent on which to base simulation space cannot be 0 length in all dimensions.");

			if (gpu == null)
				throw new ArgumentNullException("gpu");
			if (theSources == null)
				throw new ArgumentNullException("theSources");
			if (microphones == null)
				throw new ArgumentNullException("microphones");

			if (theSources.Length == 0)
				throw new ConstraintException("Sources cannot have length 0 but must have at least one element.");
			if (microphones.Length == 0)
				throw new ConstraintException("Microphones cannot have length 0 but must have at least one element.");

			_gpu = gpu;
			KindE = kindExtent;

			TheSoundField = new CSoundField(kindExtent, gpu);

			TheWalls = new Wall<byte>(kindExtent, gpu, airValue: 0);

			TheSources = theSources;

			TheMicrophones = microphones;

			_recordings = new float[_theMicrophones.Length,RecordingsInitialLength];
			_recordingsLength = RecordingsInitialLength;
			
			TheGridSettingsArray = new[] {gridSettings};

			_blocks = new dim3(
				Toolbox.RoundUpDiv(KindE.X, CudaSettings.BlockDim),
				Toolbox.RoundUpDiv(KindE.Y, CudaSettings.BlockDim));
			_threads = new dim3(CudaSettings.BlockDim, CudaSettings.BlockDim);
		}

		/// <summary>
		/// Sets the array of Sources and updates the GPU accordingly
		/// </summary>
		public Source[] TheSources
		{
			set
			{
				_theSources = value;
				UpdateToGPU(_theSources, ref _devTheSources);
			}
		}

		/// <summary>
		/// Sets the array of Microphones and updates the GPU accordingly
		/// </summary>
		public Microphone[] TheMicrophones
		{
			set
			{
				_theMicrophones = value;
				MicrophoneStruct[] micStructs = Microphone.ProduceInnerMicrophonesArray(_theMicrophones);
				UpdateToGPU(micStructs, ref _devTheMicrophoneStructs);
			}
		}

		/// <summary>
		/// Sets the GridSettingsArray and updates the GPU accordingly
		/// </summary>
		public GridSettings[] TheGridSettingsArray
		{
			set
			{
				_theGridSettingsArray = value;
				UpdateToGPU(_theGridSettingsArray, ref _devTheGridSettingsArray);
			}
		}

		public float DT
		{
			get
			{
				if (_theGridSettingsArray.Length >= 1)
					return _theGridSettingsArray[0].DT;
				throw new NotImplementedException();
			}
		}

		/// <summary>
		/// Gets the GridSettingsArray on the GPU aka device
		/// </summary>
		public GridSettings[] DevTheGridSettingsArray
		{
			get { return _devTheGridSettingsArray; }
		}

		#region IDisposable Members

		public void Dispose()
		{
			FreeGPUMemory();
		}

		#endregion

		private void UpdateToGPU<T>(T[] hostData, ref T[] devData)
		{
			if (!_gpu.IsOnGPU(devData))
				devData = _gpu.Allocate(hostData);
			_gpu.CopyToDevice(hostArray: hostData, devArray: devData);
		}

		private void UpdateToGPU(
			GridSettings[] gridSettingsArray,
			ref GridSettings[] devGridSettingsArray)
		{
			if (!_gpu.IsOnGPU(devGridSettingsArray))
				devGridSettingsArray = _gpu.Allocate(gridSettingsArray);
			_gpu.CopyToDevice(hostArray: gridSettingsArray, devArray: devGridSettingsArray);
		}

		~CSimulationSpace()
		{
			Dispose();
		}

		/// <summary>
		/// Gets the Source at the specified index
		/// </summary>
		/// <param name="index"></param>
		/// <returns></returns>
		public Source GetSource(int index)
		{
			return _theSources[index];
		}

		/// <summary>
		/// Gets the Microphone at the specified index
		/// </summary>
		/// <param name="index"></param>
		/// <returns></returns>
		public Microphone GetMicrophone(int index)
		{
			return _theMicrophones[index];
		}

		/// <summary>
		/// Copies data from a MaterialModel and uses its cells'
		/// MaterialByte getter to translate from cell material data
		/// to a byte which can be used on the GPU
		/// </summary>
		/// <param name="materialModel"></param>
		internal void SetMaterialFrom(MaterialModel materialModel)
		{
			if (materialModel == null) return;
			if (materialModel.Cells == null) return;
			foreach (Cell cell in materialModel.Cells)
			{
				if (cell.Position.X < 0 || cell.Position.X >= KindE.X)
					continue;
				if (cell.Position.Y < 0 || cell.Position.Y >= KindE.Y)
					continue;
				if (cell.Position.Z < 0 || cell.Position.Z >= KindE.Z)
					continue;
				TheWalls[cell.Position.X,
				         cell.Position.Y,
				         cell.Position.Z] = cell.MaterialByte;
				
			}
			
		}

		/// <summary>
		/// Frees up GPU memory allocated by this instance
		/// </summary>
		private void FreeGPUMemory()
		{
			foreach (object devThing in new object[]
			                            	{
			                            		_devTheSources,
			                            		_devTheMicrophoneStructs,
			                            		_devTheGridSettingsArray,
												_devAbsorption,
												_devRecordings,
			                            	})
			{
				if (_gpu.IsOnGPU(devThing))
					_gpu.Free(devThing);
			}
			//TheMaterial.FreeGPUMemory(); // find out how to destruct objects
			//TheSoundField.FreeGPUMemory();
		}

		public WallAbsorberAssociations WallAbsorberAssociations
		{
			set; get;
		}

		private float[] _devAbsorption;
		public float[] DevAbsorption
		{
			get
			{
				float[] absorption = new float[sizeof(byte)]; // 256 elements
				if (WallAbsorberAssociations != null)
				{
					for (int i = 0; i < WallAbsorberAssociations.Count; i++)
					{
						int materialByte = i + 1;
						var mapping = WallAbsorberAssociations.AbsorberMappings[i];
						if (mapping == null) continue;
						float value = mapping.AbsorberCoefficients.ToBandCoefficients().Alpha125HzBand;
						SetAbsorptionValueAt(absorption, materialByte, value);
					}
				}
				UpdateToGPU(absorption, ref _devAbsorption);
				return _devAbsorption;
			}
		}

		private const int RecordingsInitialLength = 64;
		private int _recordingsLength;
		private float[,] _devRecordings;
		private float[,] _recordings;
		public float[,] DevRecordings
		{
			get
			{
				if (!_gpu.IsOnGPU(_devRecordings))
				{
					// first time
					_devRecordings = _gpu.Allocate(_recordings);
				}
				while (Generation >= _recordingsLength)
				{
					// double array size and transfer
					int oldRecordingsLength = _recordingsLength;
					_recordingsLength *= 2;
					_recordings = new float[_theMicrophones.Length,_recordingsLength];
					float[,] devOld = _devRecordings; // temporary pointer
					_devRecordings = _gpu.Allocate(_recordings);
					_gpu.CopyOnDevice(devOld, 0, _devRecordings, 0, oldRecordingsLength);
				}
				return _devRecordings;
			}
		}

		public float[,] GetRecordings()
		{
			_gpu.CopyFromDevice(_devRecordings, _recordings); // _recordings is kept at the same size
			return _recordings;
		}

		public int RecordingsLength
		{
			get { return _recordingsLength; }
		}

		internal MicrophoneStruct[] DevTheMicrophoneStructs
		{
			get { return _devTheMicrophoneStructs; }
		}

		internal Source[] DevTheSources
		{
			get { return _devTheSources; }
		}

		/// <summary>
		/// Sets absorption value for the walls in TheWalls.
		/// <example>
		/// absorption[A] = B means that for some material with index=A-1,
		/// the absorption is B.
		/// </example>
		/// <remarks>
		/// A must be between 1 and the length of absorption 
		/// (typically 255).
		/// 
		/// TheWalls[(position)] = 0 means that there is no wall,
		/// while TheWalls[(position)] = A, (A in 1-255) means that
		/// the absorption of the wall is absorption[A].
		/// absorption[0] will not be used but is 0.
		/// </remarks>
		/// </summary>
		private void SetAbsorptionValueAt(float[] absorption, int materialByte, float value)
		{
			if (materialByte < 1 || materialByte > absorption.Length) return;

			absorption[materialByte-1] = value;
		}

		#region Generation and event
		/// <summary>
		/// Returns the current generation and subsequently increments it by one
		/// </summary>
		/// <returns></returns>
		internal int NextGeneration()
		{
			return Generation++;
		}

		// Invoke the Changed event; called whenever list changes
		public virtual void OnGenerationChanged(int generation)
		{
			if (GenerationChanged != null)
				GenerationChanged(this, generation);
		}

		private int _generation;
		public int Generation
		{
			get { return _generation; }
			set
			{
				_generation = value;
				OnGenerationChanged(_generation);
			}
		}

		// An event that clients can use to be notified whenever the
		// elements of the list change.
		public event GenerationChangedEventHandler GenerationChanged;
		#endregion

		internal void Reset()
		{
			Generation = 0;
			TheSoundField.Clear();
		}
	}
}