﻿using System;
using Cudafy.Host;
using OlaCudafy1.FDTD;

namespace OlaCudafy1.FDTD
{
	/// <summary>
	/// (Beware: does not take responsibility of deallocating its memory from the GPU.)
	/// </summary>
	internal class CSoundField : IDisposable
	{
		private float[,,] _devP;

		///// <summary>
		///// The unit of the particle speeds is:
		/////  not meters per 1 second,
		/////  but meters per [duration of simulation time step DT]
		///// </summary>
		private float[,,] _devHorizontalSpeeds;
		private float[,,] _devVerticalSpeeds;
		private float[,,] _devZSpeeds;
		private readonly KindExtent _kindExtent;
		private readonly GPGPU _gpu;

		public KindExtent KindExtent
		{
			get { return _kindExtent; }
		}

		public float[,,] DevP
		{
			get { return _devP; }
		}

		public float[,,] DevHorizontalSpeeds
		{
			get { return _devHorizontalSpeeds; }
		}

		public float[,,] DevVerticalSpeeds
		{
			get { return _devVerticalSpeeds; }
		}

		public float[,,] DevZSpeeds
		{
			get { return _devZSpeeds; }
		}

		/// <summary>
		/// Allocates memory on the GPU according to the size specified in kindExtent
		/// </summary>
		/// <param name="kindExtent"></param>
		/// <param name="gpu">The GPU on which to allocate</param>
		public CSoundField(KindExtent kindExtent, GPGPU gpu)
		{
			_kindExtent = kindExtent;
			_gpu = gpu;
			try
			{
				_devP = gpu.Allocate<float>(kindExtent.X, kindExtent.Y, kindExtent.Z);
				_devHorizontalSpeeds = gpu.Allocate<float>(kindExtent.X, kindExtent.Y, kindExtent.Z);
				_devVerticalSpeeds = gpu.Allocate<float>(kindExtent.X, kindExtent.Y, kindExtent.Z);
				_devZSpeeds = gpu.Allocate<float>(kindExtent.X, kindExtent.Y, kindExtent.Z);
			}
			catch( CudafyHostException e)
			{
				if (e.Message == "CUDA.NET exception: ErrorOutOfMemory (Ensure that memory on GPU is being explicitly released with Free()).")
					throw new OutOfMemoryException("Out of GPU memory");
				throw;
			}
		}

		/// <summary>
		/// Sets all pressure and particle velocity fields to zero.
		/// </summary>
		public void Clear()
		{
			if (_gpu.IsOnGPU(DevP))					_gpu.Set(DevP); // sets the specified device array to zero
			if (_gpu.IsOnGPU(_devHorizontalSpeeds))	_gpu.Set(_devHorizontalSpeeds);
			if (_gpu.IsOnGPU(_devVerticalSpeeds))	_gpu.Set(_devVerticalSpeeds);
			if (_gpu.IsOnGPU(_devZSpeeds))			_gpu.Set(_devZSpeeds);
		}


		internal void CopyToDevP(float[,,] p)
		{
			if (_gpu.IsOnGPU(DevP))
			{
				if (p.Length == _kindExtent.Length)
					_gpu.CopyToDevice(p, DevP);
				else
					throw new HostAndDeviceArrayDimensionsAreIncompatibleException();
			}
			else
			{
				throw new MemoryNotOnGPUOrNotInitializedException();
			}
		}

		~CSoundField()
		{
			Dispose();
		}

		public void Dispose()
		{
			FreeGPUMemory();
		}

		private void FreeGPUMemory()
		{
			foreach (var devThing in new object[]
			                         	{
			                         		_devP,
											_devHorizontalSpeeds,
											_devVerticalSpeeds,
											_devZSpeeds
			                         	})
			{
				if (_gpu.IsOnGPU(devThing))
					_gpu.Free(devThing);
			}
		}
	}
}