﻿using System;
using Cudafy;
using Cudafy.Host;

namespace OlaCudafy1.FDTD
{
	/// <summary>
	/// Generates for itself the data to draw from
	/// </summary>
	internal class DumbDraw
	{
		private const int Four = 4;

		private static readonly Type[] Types = new[]
		                                       	{
		                                       		typeof (MicrophoneStruct),
		                                       		typeof (Source),
		                                       		typeof (GridSettings),
		                                       		typeof (KindExtent),
		                                       		typeof (DumbDraw)
		                                       	};

		private readonly GPGPU _gpu;

		private byte[,,] _destinationImage;
		private byte[,,] _devDestinationImage;
		private readonly CSimulationSpace _simulationSpace;
		private readonly DisplaySettings _displaySettings;
		private int _imageW;
		private int _iHeight;

		public DumbDraw(CSimulationSpace simulationSpace, DisplaySettings displaySettings, GPGPU gpu)
		{
			_gpu = gpu;
			_simulationSpace = simulationSpace;
			_displaySettings = displaySettings;
			_displaySettings.Changed += DisplaySettingsChanged;
			DisplaySettingsChanged();
		}

		private void DisplaySettingsChanged()
		{
			switch (_displaySettings.SliceDimension)
			{
				case SliceDimension.X: // x
					_imageW = _simulationSpace.KindE.Y;
					_iHeight = _simulationSpace.KindE.Z;
					break;
				case SliceDimension.Y: // y
					_imageW = _simulationSpace.KindE.Z;
					_iHeight = _simulationSpace.KindE.X;
					break;
				case SliceDimension.Z: // z (2)
					_imageW = _simulationSpace.KindE.X;
					_iHeight = _simulationSpace.KindE.Y;
					break;
				default:
					_imageW = 0; // TODO test!
					_iHeight = 0;
					break;
			}
		}

		internal void Launch()
		{
			lock (this)
			{
				// load the kernel module and dependencies (structs etc)
				_gpu.LoadModule(SuperCudafier.Cudafy(Types));

				// allocate space for image in host and in device
				_destinationImage = new byte[_imageW, _iHeight, Four];
				_devDestinationImage = _gpu.Allocate<byte>(_imageW, _iHeight, Four);

				// generate p conents
				var p = new float[
					_simulationSpace.KindE.X,
					_simulationSpace.KindE.Y,
					_simulationSpace.KindE.Z];
				for (int i = 0; i < _simulationSpace.KindE.X; i++)
				{
					for (int j = 0; j < _simulationSpace.KindE.Y; j++)
					{
						for (int k = 0; k < _simulationSpace.KindE.Z; k++)
						{
							p[i, j, k] = i + j + k;
						}
					}
				}
				var zero = new float[
					_simulationSpace.KindE.X,
					_simulationSpace.KindE.Y,
					_simulationSpace.KindE.Z];

				var sim = _simulationSpace;
				var disp = _displaySettings;

				// copy the generated pressure field p to the dev mem
				_gpu.CopyToDevice(p, sim.TheSoundField.DevP);
				_gpu.CopyToDevice(zero, sim.TheSoundField.DevHorizontalSpeeds);
				_gpu.CopyToDevice(zero, sim.TheSoundField.DevVerticalSpeeds);
				_gpu.CopyToDevice(zero, sim.TheSoundField.DevZSpeeds);

				var blocks = new dim3(
					Toolbox.RoundUpDiv(_imageW, CudaSettings.BlockDim),
					Toolbox.RoundUpDiv(_iHeight, CudaSettings.BlockDim));
				var threads = new dim3(CudaSettings.BlockDim, CudaSettings.BlockDim);
				_gpu.Launch(blocks, threads).DumbDrawKernel(
					sim.TheSoundField.DevP,
					sim.TheSoundField.DevHorizontalSpeeds,
					sim.TheSoundField.DevVerticalSpeeds,
					sim.TheSoundField.DevZSpeeds,
					_devDestinationImage,
					sim.TheWalls.DevMaterialData,
					scalePressure: disp.ScalePressure,
					scaleBoth: disp.ScaleBoth,
					displaySlice: disp.DisplaySlice,
					sliceDimension: disp.SliceDimensionInt,
					imgEx: new KindExtent {X = _imageW, Y = _iHeight, Z = 1}
					);

				CopyToHost();

				_gpu.Free(_devDestinationImage);
			}
		}

		[Cudafy]
		public static void DumbDrawKernel(GThread thread,
		                                  float[,,] p,
		                                  float[,,] horiz,
		                                  float[,,] vert,
		                                  float[,,] zData,
		                                  byte[,,] dest,
										  byte[, ,] mat,
										  float scalePressure,
		                                  float scaleBoth,
		                                  int displaySlice,
		                                  int sliceDimension,
		                                  KindExtent imgEx
			)
		{
			// Calculate the indexes in the image for the thread
			int w = thread.threadIdx.x + thread.blockIdx.x*thread.blockDim.x;
			int h = thread.threadIdx.y + thread.blockIdx.y*thread.blockDim.y;

			if (w >= imgEx.X || h >= imgEx.Y) return;
			if (w < 0 || h < 0) return;

			dest[w, h, 0] = (byte)
			                (((w + h)/(8F))*255F);
			dest[w, h, 1] = 0;
			dest[w, h, 2] = 0;
			dest[w, h, 3] = 255; // alpha
		}



		private void CopyToHost()
		{
			_gpu.CopyFromDevice(_devDestinationImage, _destinationImage);
		}

		public byte[, ,] DestinationImage
		{
			get { return _destinationImage; }
		}

	}
}