﻿/*
 * Copyright 2009-2010 Ola Vikholt.  All rights reserved.
 *
 * Ola Vikholt retains all intellectual property and
 * proprietary rights in and to this software and related documentation and
 * any modifications thereto.  Any use, reproduction, disclosure, or distribution
 * of this software and related documentation without an express license
 * agreement from Ola Vikholt is strictly prohibited.
 *
 */


// Ah, alt+' er bare vanvittig bra

using Cudafy;
using Cudafy.Host;
using OlaCudafy1.FDTD;
using OlaCudafy1.Material;

namespace OlaCudafy1.FDTD
{

	public class FDTDCalcGpu : ICalc
	{
		
		internal FDTDCalcGpu(CSimulationSpace simSpace, GPGPU gpu, SimConfig simConfig)
		{
			
			_gpu = gpu;
			_simConfig = simConfig;

			_simulationSpace = simSpace;
		}

		private static readonly System.Type[] Types = new[]
		                                            {
		                                                typeof (KindExtent),
		                                                typeof (Point),
		                                                typeof (MicrophoneStruct),
		                                                typeof (Source),
		                                                typeof (GridSettings),
		                                                typeof (FDTDCalcGpu)
		                                            };

		private readonly CSimulationSpace _simulationSpace;

		private readonly GPGPU _gpu;
		private readonly SimConfig _simConfig;
		
		private int _generation;

		private float _elapsedTimeInMilliseconds;

		public float ElapsedTimeInMilliseconds
		{
			get { return _elapsedTimeInMilliseconds; }
		}

		
		// An event that clients can use to be notified whenever the
		// elements of the list change.
		public event GenerationChangedEventHandler GenerationChanged;

		// Invoke the Changed event; called whenever list changes
		public virtual void OnGenerationChanged(int generation)
		{
			if (GenerationChanged != null)
				GenerationChanged(this, generation);
		}

		public int Generation
		{
			get { return _generation; }
			set {
				_generation = value;
				OnGenerationChanged(_generation);
			}
		}

		public void Execute()
		{
			_gpu.LoadModule(SuperCudafier.Cudafy(Types));

			#region help resolve errors
			// === Kernel ===
			// Errors reported from (seems to originate) here (*) and their solution (->):
			// *CudafyHostException: "Could not find function ..."
			//   -> The kernel must be marked as public and not private nor internal
			// *ArgumentException: "Argument 1 of type 
			//  (your namespace).(your struct containing another struct)[] (eg. Lek1.Cairn[])
			//  is not on the GPU or not supported."
			//   -> Sadly, Cudafy doesn't seem to support nested types
			//   -> ???
			// *ArgumentException: "Object contains non-primitive or non-blittable data."
			//   -> C# did not manage to pass your argument to the kernel because your struct
			//      contained another struct, or a bool.
			//      Bool is forbidden! +_+
			// -----
			#endregion
			_gpu.StartTimer();
			_gpu.Launch(_simulationSpace._blocks, _simulationSpace._threads).CalcKernel(
				_simulationSpace.DevTheGridSettingsArray,
				_simConfig.DoMur ? 1 : 0,
				_simulationSpace.TheSoundField.DevP,
				_simulationSpace.TheSoundField.DevHorizontalSpeeds,
				_simulationSpace.TheSoundField.DevVerticalSpeeds,
				_simulationSpace.TheSoundField.DevZSpeeds,
				_simulationSpace.TheWalls.DevMaterialData,
				_simulationSpace.KindE,
				_simulationSpace.DevTheSources,
				_simulationSpace.DevTheMicrophoneStructs,
				Generation // uint max => max simulation time ≈ 12 hours
				); // CalcKernel() // rename here also

			_elapsedTimeInMilliseconds = _gpu.StopTimer();
			++Generation;
		}

		/// <summary>
		/// Calculates one step of Vikholt's reimplementation 
		/// and extension of the Sakamoto laboratory's stagged-grid FDTD algorithm
		/// </summary>
		[Cudafy]
		public static void CalcKernel(GThread thread,
			GridSettings[] devGridAr,
			int doMur,
			float[, ,] p,
			float[, ,] horiz,
			float[, ,] vert,
			float[, ,] zData,
			byte[, ,] mat,
			KindExtent ex,
			Source[] srcs,
			MicrophoneStruct[] mics,
			int generation // uint max => max simulation time ≈ 12 hours
			)
		{
			// ReSharper disable InconsistentNaming
			GridSettings grid = devGridAr[0];

			// translate from thread index and block index to pixel position
			int x = thread.threadIdx.x + thread.blockIdx.x*thread.blockDim.x;
			int y = thread.threadIdx.y + thread.blockIdx.y*thread.blockDim.y;

			if (x >= ex.X || y >= ex.Y)
				return;

			const byte W = 0x7;
			int z;
			// =========== v, u, w ============

			for (z = 0; z < ex.Z; ++z)
			{
				var isWallHere = (byte)(mat[x, y, z] & W);
				if (isWallHere == 0 // there's NO wall at the source (to the left) of this particle speed
				    || (x == 0 || y == 0 || (z == 0)) // disregard material at top borders
					)
				{
					// if all except outer border, and no wall
					if (x < ex.X - 1 && (mat[x + 1, y, z] & W)==0)
						horiz[x, y, z] += grid.CU*(
						                         	((x > 0 ? 1F : 0F)
													*p[x, y, z]
							//                       ^^^^^^^^^^^^^^^^
							// the above line: ignore pressure at 0-indices (i.e. at mem padding)
													- p[x + 1, y, z])
													);
					else if (x == ex.X - 1) // at last point in grid aka at outer border
						horiz[x, y, z] += grid.CU*p[x, y, z]; // particle speed outward (in positive direction) only
					if (y < ex.Y - 1 && (mat[x, y + 1, z] & W) == 0)
						vert[x, y, z] += grid.CU*(
						                         	((y > 0) ? 1F : 0F)*p[x, y, z]
													- p[x, y + 1, z]
						                         );
					else if (y == ex.Y - 1)
						vert[x, y, z] += grid.CU * p[x, y, z];
					if (z < ex.Z - 1 && (mat[x, y, z + 1] & W) == 0)
						zData[x, y, z] += grid.CU*(
							(z > 0 ? 1F : 0F)
							*p[x, y, z]
							- p[x, y, z + 1]);
					else if (z == ex.Z - 1)
						zData[x, y, z] += grid.CU*p[x, y, z];
				}
				else
				{
					// WALL_HERE
					const float alpha = 1F; //0.2;
					float ZN = grid.R0*grid.C0*
					           (1F + GMath.Sqrt(1F - alpha))/(1F - GMath.Sqrt(1F - alpha));
				}
			}
			if (doMur != 0)
			{
				// mur after
				for (z = 0; z < ex.Z; ++z)
				{
					if (x == 0)
						horiz[x, y, z] += grid.MurCoefficient3*horiz[x + 1, y, z];
					if (y == 0)
						vert[x, y, z] += grid.MurCoefficient3*vert[x, y + 1, z];
					if (z == 0)
						zData[x, y, z] += grid.MurCoefficient3*zData[x, y, z + 1];
					if (x == ex.X - 1)
						horiz[x, y, z] += grid.MurCoefficient3*horiz[x - 1, y, z];
					if (y == ex.Y - 1)
						vert[x, y, z] += grid.MurCoefficient3*vert[x, y - 1, z];
					if (z == ex.Z - 1)
						zData[x, y, z] += grid.MurCoefficient3*zData[x, y, z - 1];
				}
			}

			thread.SyncThreads();
			// =========== p ============

			for (z = 0; z < ex.Z; ++z)
			{
				if ((mat[x, y, z] & W) == 0)
				{
					if (x > 0 && y > 0 && z > 0)
					{
						p[x, y, z] += grid.CP*(horiz[x - 1, y, z] - horiz[x, y, z]);
						p[x, y, z] += grid.CP*(vert[x, y - 1, z] - vert[x, y, z]);
						p[x, y, z] += grid.CP*(zData[x, y, z - 1] - zData[x, y, z]);
					}
				}
				else
				{
					// vanlig:
					p[x, y, z] = 0F;
					// impedance wall test:
					//float S = grid.DX;
					//const float k = .8F;
					//float C = grid.R0*grid.C0*S; // <-
					//const float m = 0.00F;
					//float denominator = m + C*grid.DT + k*grid.DT*grid.DT;
					//float konst = 2*C*grid.DT + k*grid.DT*grid.DT;
					//float dxdt = horiz[x, y, z];
					//p[x, y, z] += dxdt;
					//dxdt = (dxdt*m - konst*p[x, y, z] - p[x + 1, y, z]*S*grid.DT)/denominator;
					//horiz[x, y, z] = dxdt;
				}

				// sources
				foreach (var src in srcs)
				{
					if (x == src.X && y == src.Y && z == src.Z)
					{
						float time = generation*grid.DT;
						p[x, y, z] = src.PressureAmplitude*GMath.Sin(src.Frequency*2*GMath.PI*time);
					}

					// initial field
					 //if (generation == 0)
					 //{
					 //   int mx = (int) src.X;
					 //   int my = (int) src.Y;

					 //    // choose from two location methods:
					 //    // between pressure grid cells (more smooth)
					 //   float dx = (x - mx + 0.5) * grid.DX;
					 //    float dy = (y - my + 0.5) * grid.DX;
					 //    // on single pressure grid cell (more precise / pin-pointed; same spot as sweep source)
					 //    //		float dx = (ix-mx)*DX;
					 //    //		float dy = (iy-my)*DX;
					 //    // the same goes for the 3D version below, which is not updated

					 //    float spikeBigness = 0.002f;
					 //    p[x,y,z] = sourcePressureAmplitude * GMath.Exp(-(
					 //                            dx * dx +
					 //                            dy * dy
					 //                            ) / spikeBigness);

					 //    //		float shikouseiAngle = -0.75*PI;
					 //    //		U_RIGHT =     __cosf(shikouseiAngle)*P_HERE/IMPEDANCE;
					 //    //		U_DOWN = __sinf(shikouseiAngle)*P_HERE/IMPEDANCE;

					 //}

					if (generation == src.InvokeAtGeneration)
					{
						const int nx = 6; // number of wavelengths in window
						const int ny = 6; // number of wavelengths in window
						float kx = nx * 2F * GMath.PI / grid.DX / (ex.X); // wave number aka spatial freq
						float ky = ny * 2F * GMath.PI / grid.DX / (ex.Y);
						p[x, y, z] = srcs[0].PressureAmplitude *
									 GMath.Sin(kx * x * grid.DX) * GMath.Sin(ky * y * grid.DX);
						//horiz[x, y, z] = srcs[0].PressureAmplitude *
						//GMath.Sin(kx * x * grid.DX) * GMath.Sin(ky * y * grid.DX) / grid.Impedance;
						p[x, y, z] = srcs[0].PressureAmplitude;
						//10 * GMath.Sin(kx * x * grid.DX) * GMath.Sin(ky * y * grid.DX); ;
						// srcs[0].PressureAmplitude;
					}
				}


				//p[x, y, z] = x * y * z + generation;
				//p[x, y, z] = (float)generation;

				
			}
			// ReSharper restore InconsistentNaming
		}
	}
}



//        // --- impulse source ---
//        // (needs to be inside of the for loop)
//        impulseSource(generation,
//                  sourceType,
//                  sourceLocation,
//                  pData,
//                  horizontal, verticalSpeeds,
//                  lin,
//                  ix, iy, z,
//                  dims,
//                  DX,
//                  sourcePressureAmplitude,
//                  isVolume);


//        // horn source
//        if (
//            sourceType == None &&
//            mat[matlin] == Material::MaterialTypeSource &&
//            ix > 0 && iy > 0 && ix<dims.x-1 && iy < dims.y-1) {
//            for (int sourceIndex = 0; sourceIndex < sourceStructArrayLength; ++sourceIndex) {
//                if (z == sourceStructArrayOnDevice[sourceIndex].dims.z
//                        && (ix == sourceStructArrayOnDevice[sourceIndex].dims.x)
//                        && (iy == sourceStructArrayOnDevice[sourceIndex].dims.y)) {
//                    if (sourceStructArrayOnDevice[sourceIndex].sourceType == Velocity) {
//                        float t = generation*DT;
//                        //float frequency = __expf(timeScale*t+timeOffset)+frequencyOffset;
////							float3 velocity = make_float3(0.001, 0.0, 0.0);
//                        float3 velocity = sourceStructArrayOnDevice[sourceIndex].velocity;
//                        U_RIGHT = velocity.x * sinf(frequency*2*M_PI*t);
//                        U_DOWN = velocity.y * sinf(frequency*2*M_PI*t);
//                        U_IN = velocity.z * sinf(frequency*2*M_PI*t);
//                    }
//                }
//            }
//        }

//    }

//    // --- sources ---
//    int boomPos = 0;
//    if (doSonicBoom) {
//        boomPos = generation/BOOM;
//        if (boomPos >= dims.x)
//            boomPos -= dims.x;
//    }
//    int zlsize = 0;
//    //		if (isVolume)
//    //			zlsize = 4;
//    for (int zl = sourceLocation.z-zlsize; zl <= sourceLocation.z+zlsize; ++zl) {

//        // point type
//        if (ix == sourceLocation.x+boomPos && (iy <= sourceLocation.y+zlsize && iy >= sourceLocation.y-zlsize)) {
//            // IN_PROGRESS temp plane source
//            lin = zl*realSlicePitch + iy*realPitch + ix;
//            float time = generation*DT;
//            if (sourceType == Sine) {
//                // source
//                P_HERE = sourcePressureAmplitude * sinf(sourceFrequency*2*M_PI*time); //*0.02f*20;
//            } else if (sourceType == Sweep) {
//                // source
//                P_HERE = sourcePressureAmplitude * sinf(frequency*2*M_PI*time); //*0.02f*20; // hard source (=)
////				P_HERE += sourcePressureAmplitude * sinf(frequency*2*M_PI*time); // soft source (+=)
//            } /*else if (sourceType == Spk && generation < spkLen) {
//             // mic source
//             //P_HERE += spkData[generation]; // soft source (+=)
//             P_HERE = spkData[generation];
//            }*/
//        }
//    }

//    // --- line source ---
//    if (sourceType == LinePulse && ix == sourceLocation.x) {
//        lin = sourceLocation.z*realSlicePitch + iy*realPitch + ix;
//        float t = generation*DT;
//        float f = frequency;//sourceFrequency; // 5000.0;
//        P_HERE = sourcePressureAmplitude * sinf(f*2*M_PI*t)*0.02f*50; //BINARY(__sinf(f*2*M_PI*t)*0.02f*50);
//    }

////	if (mat[matlin] == Material::MaterialTypeMicrophone) {
//    // -- mic --
//    if (ix == micLocation.x && iy == micLocation.y && generation < MICDATALEN) {
//        lin = micLocation.z*realSlicePitch + iy*realPitch + ix;
////		micData[generation] = P_HERE; // pressure mic
//        *micSampleOnDevice = P_HERE;
//    }
////	}

//} // _FDTD_calc_kernel


//__device__ void inline impulseSource2D(
//    unsigned int generation,
//    SourceType sourceType,
//    int3 sourceLocation,
//    Real *pData,
//    Real *horizontal, Real *verticalSpeeds,
//    int lin,
//    int ix, int iy,
//    int3 dims,
//    Real DX,
//    Real sourcePressureAmplitude)
//{
//    if (generation == 0 && sourceType == Impulse) {
//        int mx = (int)sourceLocation.x;
//        int my = (int)sourceLocation.y;

//        // choose from two location methods:
//        // between pressure grid cells (more smooth)
//        float dx = (ix-mx+0.5)*DX;
//        float dy = (iy-my+0.5)*DX;
//        // on single pressure grid cell (more precise / pin-pointed; same spot as sweep source)
////		float dx = (ix-mx)*DX;
////		float dy = (iy-my)*DX;
//        // the same goes for the 3D version below, which is not updated

//        float spikeBigness = 0.002f;
//        P_HERE = sourcePressureAmplitude * expf( -(
//                                dx*dx+
//                                dy*dy
//                                )/spikeBigness );

////		float shikouseiAngle = -0.75*PI;
////		U_RIGHT =     __cosf(shikouseiAngle)*P_HERE/IMPEDANCE;
////		U_DOWN = __sinf(shikouseiAngle)*P_HERE/IMPEDANCE;

//    } else if (generation == 0 && sourceType == Field) {
//        int nx = 6, ny = 6; // number of wavelengths in window
//        float kx = nx*2.0*PI/DX/(dims.x); // wave number aka spatial freq
//        float ky = ny*2.0*PI/DX/(dims.y);
//        P_HERE = sourcePressureAmplitude *
//                sinf(kx * ix * DX) * sinf(ky * iy * DX);
//        U_RIGHT = sourcePressureAmplitude *
//                sinf(kx * ix * DX) * sinf(ky * iy * DX) / IMPEDANCE;
////		U_DOWN
////		P_HERE = sourcePressureAmplitude * 0.5*
////				(sinf(k * ix * DX) + __sinf(k * iy * DX));
//    }
//}

//__device__ void inline impulseSource3D(
//    unsigned int generation,
//    SourceType sourceType,
//    int3 sourceLocation,
//    Real *pData, int lin,
//    int ix, int iy, int z,
//    int3 dims,
//    Real DX,
//    Real sourcePressureAmplitude)
//{
//    if (generation == 0 && sourceType == Impulse) {
//        int mx = (int)sourceLocation.x;
//        int my = (int)sourceLocation.y;
//        int mz = (int)sourceLocation.z;
//        float spikeBigness = 0.002f;
//        P_HERE =
//            sourcePressureAmplitude *
//            expf( -(
//                (ix-mx-0.5)*DX*(ix-mx-0.5)*DX+
//                (iy-my-0.5)*DX*(iy-my-0.5)*DX+
//                ( z-mz-0.5)*DX*( z-mz-0.5)*DX
//                )/spikeBigness );
//    }
//}

//__device__ void inline impulseSource(
//        unsigned int generation,
//        SourceType sourceType,
//        int3 sourceLocation,
//        Real *pData, Real *horizontal, Real *verticalSpeeds,
//        int lin,
//        int ix, int iy, int z,
//        int3 dims,
//        Real DX,
//        Real sourcePressureAmplitude,
//        bool isVolume)
//{
//    if (isVolume) {
//        impulseSource3D(generation,
//                sourceType,
//                sourceLocation,
//                pData, lin,
//                ix, iy, z,
//                dims,
//                DX,
//                sourcePressureAmplitude);
//    } else {
//        impulseSource2D(generation,
//                sourceType,
//                sourceLocation,
//                pData,
//                horizontal, verticalSpeeds,
//                lin,
//                ix, iy,
//                dims,
//                DX,
//                sourcePressureAmplitude);
//    }
//}


//__device__ void inline murBefore(
//        Real *horizontal,
//        Real *verticalSpeeds,
//        Real *wData,
//        int ix, int iy, bool isVolume,
//        int3 dims,
//        int realPitch, int realSlicePitch,
//        Real MurCoefficient1, Real MurCoefficient2)
//{
//    int lin = 0;
//    for (int z = 0; z < dims.z; ++z) {
//        lin = z*realSlicePitch + iy*realPitch + ix;
//        if (ix == 0)
//            U_RIGHT = MurCoefficient1*U_RIGHT_RIGHT-MurCoefficient2*U_RIGHT;
//        if (iy == 0)
//            U_DOWN = MurCoefficient1*U_DOWN_DOWN-MurCoefficient2*U_DOWN;
//        if ((isVolume && z == 0))
//            U_IN = MurCoefficient1*U_IN_IN-MurCoefficient2*U_IN;

//        if (ix == dims.x-1)
//            U_RIGHT = MurCoefficient1*U_LEFT-MurCoefficient2*U_RIGHT;
//        if (iy == dims.y-1)
//            U_DOWN = MurCoefficient1*U_UP-MurCoefficient2*U_DOWN;
//        if ((isVolume && z == dims.z-1))
//            U_IN = MurCoefficient1*U_OUT-MurCoefficient2*U_IN;
//    }
//}












//var replaceList = new Dictionary<string, string>();
			//replaceList["DevPLen1"] = "theSimulationSpace.KindE.Axes[0]";
			//replaceList["DevPLen2"] = "theSimulationSpace.KindE.Axes[1]";
			//replaceList[GetName(P_HERE)] = P_HERE;
			//replaceList[GetName(P_L)] = P_L;
			//replaceList[GetName(P_R)] = P_R;
			//replaceList[GetName(P_U)] = P_U;
			//replaceList[GetName(P_D)] = P_D;
			//replaceList[GetName(P_A)] = P_A;
			//replaceList[GetName(P_Z)] = P_Z;
			//replaceList[GetName(U_RIGHT)] = U_RIGHT;
			//replaceList[GetName(U_DOWN)] = U_DOWN;
			//replaceList[GetName(U_IN)] = U_IN;
			//replaceList[GetName(U_LEFT)] = U_LEFT;
			//replaceList[GetName(U_UP)] = U_UP;
			//replaceList[GetName(U_OUT)] = U_OUT;
			//replaceList[GetName(U_RIGHT_RIGHT)] = U_RIGHT_RIGHT;
			//replaceList[GetName(U_DOWN_DOWN)] = U_DOWN_DOWN;
			//replaceList[GetName(U_IN_IN)] = U_IN_IN;
			//replaceList[GetName(WALL_HERE)] = WALL_HERE;
			//replaceList[GetName(WALL_L)] = WALL_L;
			//replaceList[GetName(WALL_R)] = WALL_R;
			//replaceList[GetName(WALL_U)] = WALL_U;
			//replaceList[GetName(WALL_D)] = WALL_D;
			//replaceList[GetName(WALL_A)] = WALL_A;
			//replaceList[GetName(WALL_Z)] = WALL_Z;




////// new field access convenience macros
//private const string P_HERE = "(pData[lin])";

//private const string P_L = "(pData[lin-1])"; // to the left
//private const string P_R = "(pData[lin+1])"; // to the right
//private const string P_U = "(pData[lin-realPitch])"; // up in x direction
//private const string P_D = "(pData[lin+realPitch])"; // down in x direction
//private const string P_A = "(pData[lin-realSlicePitch])";
//    // up in z direction (eg 'a' to fly in Duke Nukem 3D)

//private const string P_Z = "(pData[lin+realSlicePitch])";
//    // down in z direction (eg 'z' to descend in Duke Nukem 3D)

//private const string U_RIGHT = "(horizontal[lin])";
//private const string U_DOWN = "(verticalSpeeds[lin])";
//private const string U_IN = "(wData[lin])";
//    // uses right hand rule for naming: positive index change directions are right-down-in (equiv for )
//private const string U_LEFT = "(horizontal[lin-1])";
//private const string U_UP = "(verticalSpeeds[lin-realPitch])";
//private const string U_OUT = "(wData[lin-realSlicePitch])";

//private const string U_RIGHT_RIGHT= " (horizontal[lin+1])";
//private const string U_DOWN_DOWN = "(verticalSpeeds[lin+realPitch])";
//private const string U_IN_IN = "(wData[lin+realSlicePitch])";

//private const string WALL_HERE = "(mat[matlin] == Material::MaterialTypeWall)";
//private const string WALL_L = " (mat[matlin-1] == Material::MaterialTypeWall)";
//private const string WALL_R = " (mat[matlin+1] == Material::MaterialTypeWall)";
//private const string WALL_U = " (mat[matlin-charPitch] == Material::MaterialTypeWall)";
//private const string WALL_D = "(mat[matlin+charPitch] == Material::MaterialTypeWall)";
//private const string WALL_A = "(mat[matlin-charSlicePitch] == Material::MaterialTypeWall)";
//private const string WALL_Z= " (mat[matlin+charSlicePitch] == Material::MaterialTypeWall)";

//static string GetName<T>(T item) where T : class
//{
//    return typeof(T).GetProperties()[0].Name;
//}
