﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using Cudafy;
using Cudafy.Host;
using OlaCudafy1.FDTD;
using Point = System.Drawing.Point;

namespace OlaCudafy1.FDTD
{

	internal class FDTDDrawMachine
	{
		public FDTDDrawMachine(CSimulationSpace simulationSpace, DisplaySettings displaySettings, GPGPU gpu)
		{
			_gpu = gpu;
			_simulationSpace = simulationSpace;
			_displaySettings = displaySettings;
			_displaySettings.Changed += DisplaySettingsChanged;
			DisplaySettingsChanged();
			_thisModule = SuperCudafier.Cudafy(Types);
		}
		private readonly CudafyModule _thisModule;

		private void DisplaySettingsChanged()
		{
			switch (_displaySettings.SliceDimension)
			{
				case SliceDimension.X: // x
					_imageW = _simulationSpace.KindE.Y;
					_iHeight = _simulationSpace.KindE.Z;
					break;
				case SliceDimension.Y: // y
					_imageW = _simulationSpace.KindE.Z;
					_iHeight = _simulationSpace.KindE.X;
					break;
				case SliceDimension.Z: // z (2)
					_imageW = _simulationSpace.KindE.X;
					_iHeight = _simulationSpace.KindE.Y;
					break;
				default:
					_imageW = 0; // TODO test!
					_iHeight = 0;
					break;
			}
		}

		private readonly GPGPU _gpu;
		private readonly CSimulationSpace _simulationSpace;
		private readonly DisplaySettings _displaySettings;
		private byte[,,] _devDestinationImage;

		private static readonly Type[] Types = new[] {
			    typeof (MicrophoneStruct),
			    typeof (Source),
			    typeof (GridSettings),
				typeof (KindExtent),
			    typeof (FDTDDrawMachine)
			};

		int _imageW;
		int _iHeight;

		internal void Launch()
		{
				if (_thisModule == null) return;
				_gpu.LoadModule(_thisModule); // doing this every time is pretty slow

				// not thread safe! bleh

				_devDestinationImage = _gpu.Allocate<byte>(_imageW, _iHeight, Four);

				var blocks = new dim3(
					Toolbox.RoundUpDiv(_imageW, CudaSettings.BlockDim),
					Toolbox.RoundUpDiv(_iHeight, CudaSettings.BlockDim));
				var threads = new dim3(CudaSettings.BlockDim, CudaSettings.BlockDim);

				DisplaySettings disp = _displaySettings;
				CSimulationSpace sim = _simulationSpace;
				int slice = disp.DisplaySlice;
				int slcDim = disp.SliceDimensionInt;
				_gpu.Launch(blocks, threads).FDTDDrawKernel(
					sim.TheSoundField.DevP,
					sim.TheSoundField.DevHorizontalSpeeds,
					sim.TheSoundField.DevVerticalSpeeds,
					sim.TheSoundField.DevZSpeeds,
					_devDestinationImage,
					sim.TheWalls.DevMaterialData,
					scalePressure: disp.ScalePressure,
					scaleBoth: disp.ScaleBoth,
					srcs: sim.DevTheSources,
					mics: sim.DevTheMicrophoneStructs,
					displaySlice: slice,
					sliceDimension: slcDim,
					imgEx: new KindExtent() { X = _imageW, Y = _iHeight, Z = 1 }
					);

				
				var destinationImage = new byte[_imageW, _iHeight, Four];
				_gpu.CopyFromDevice<byte>(_devDestinationImage, destinationImage);

				_bmp = ComputeBitmap(destinationImage, _imageW, _iHeight);
				_gpu.Free(_devDestinationImage);

		}


		private static Bitmap ComputeBitmap(byte[,,] fromImage, int width, int height)
		{
			int byteCount = width * height * Four;

			var temp = new byte[byteCount];
			for (int x = 0; x < width; x++)
			{
				for (int y = 0; y < height; y++)
				{
					int offset = x + y * width;
					temp[offset * Four + 0] = fromImage[x, y, 2]; //b
					temp[offset * Four + 1] = fromImage[x, y, 1]; //g
					temp[offset * Four + 2] = fromImage[x, y, 0]; //r
					temp[offset * Four + 3] = fromImage[x, y, 3]; //a
				}
			}

			// bitmap to be copied to
			var bmp = new Bitmap(width, height, PixelFormat.Format32bppArgb);

			BitmapData bmpData = bmp.LockBits(new Rectangle(new Point(0, 0), bmp.Size),
											  ImageLockMode.WriteOnly, bmp.PixelFormat);
			IntPtr ptr = bmpData.Scan0;
			System.Runtime.InteropServices.Marshal.Copy(temp, 0, ptr, byteCount);
			bmp.UnlockBits(bmpData);
			
			return bmp;
		}

		private const int Four = 4;

		private Bitmap _bmp;

		// todo replace all this with opengl cuda interop
		public Bitmap Bitmap
		{
			get
			{
				return _bmp;
			}
		}




		[Cudafy]
		public static int Clip(ref float v)
		{
			if (v < 0)
				v = 0;
			else if (v > 1)
				v = 1;

			return 0; // returning void does not work, due to bug in Cudafy v1.5
		}

		/// <summary>
		/// Converts HSV [0F,1F) to RGB [0,255] -- R, G and B values are stored in out-arguments. (Please ignore return value).
		/// </summary>
		/// <returns>0. Ignore this return value.</returns>
		[Cudafy]
		public static int SetHSV(float h, float s, float v,
			out byte outR,
			out byte outG,
			out byte outB
			)
		{
			float r = 0F, g = 0F, b = 0F;
			if (s == 0F)
			{
				r = g = b = v;
			}
			else
			{
				if (h == 1F) h = 0;
				float z = GMath.Floor(h * 6);
				int i = (int)z;
				float f = h * 6 - z;
				float p = v * (1 - s);
				float q = v * (1 - s * f);
				float t = v * (1 - s * (1 - f));

				switch (i)
				{
					case 0: r = v; g = t; b = p; break;
					case 1: r = q; g = v; b = p; break;
					case 2: r = p; g = v; b = t; break;
					case 3: r = p; g = q; b = v; break;
					case 4: r = t; g = p; b = v; break;
					case 5: r = v; g = p; b = q; break;
				}
			}
			int c;
			c = (int)(256 * r); if (c > 255) c = 255; outR = (byte)c;
			c = (int)(256 * g); if (c > 255) c = 255; outG = (byte)c;
			c = (int)(256 * b); if (c > 255) c = 255; outB = (byte)c;

			return 0; // returning void does not work, due to bug in Cudafy v1.5
		}


		[Cudafy]
		// must be 'void' so that it turns __global__ (due to bug in Cudafy)
		public static void FDTDDrawKernel(GThread thread, // keep kernel naming from qtsaka
			//FDTDGridSpecStruct gridSpecStruct,
			float[, ,] p,
			float[, ,] horiz,
			float[, ,] vert,
			float[, ,] zData,
			byte[, ,] dest,
			byte[, ,] mat,

			// visualization/display customizations:
			float scalePressure,
			float scaleBoth,
			// mics and sources description and container object
			Source[] srcs,
			MicrophoneStruct[] mics,
			// how to slice -object:
			// (slice to display)
			int displaySlice,
			// (dimension to slice through)
			int sliceDimension,
			KindExtent imgEx
			)
		{
			// Calculate the indexes in the image for the thread
			int w = thread.threadIdx.x + thread.blockIdx.x * thread.blockDim.x;
			int h = thread.threadIdx.y + thread.blockIdx.y * thread.blockDim.y;
			int d = displaySlice;

			if (w >= imgEx.X || h >= imgEx.Y) return;

			int x, y, z;
			switch (sliceDimension)
			{
				case 0: // x
					x = d;
					y = w;
					z = h;
					break;
				case 1: // y
					x = h;
					y = d;
					z = w;
					break;
				default: // z (sliceDimension==2)
					x = w;
					y = h;
					z = d;
					break;
			}

			if (x < 0 || y < 0) return;
				// for drawing also the padding pressures
				// x < gridSpecStruct.GetX() &&  y < gridSpecStruct.GetY()) {


			float leftRightParticleSpeedAverage; // naming relative to axes in produced picture
			var goMinusW = (byte)(0 < w ? 1 : 0);// <=> 0<x when slice is z, 0<y when slice is x etc.
			switch (sliceDimension)
			{
				case 0: // x
					leftRightParticleSpeedAverage = (vert[x, y, z] + vert[x, y-goMinusW, z]) / 2.0F; //x
					break;
				case 1: // y
					leftRightParticleSpeedAverage = (zData[x, y, z] + zData[x, y, z-goMinusW]) / 2.0F; //y
					break;
				default: // z
					leftRightParticleSpeedAverage = (horiz[x, y, z] + horiz[x-goMinusW, y, z]) / 2.0F; //z
					break;
			}

			float upDownParticleSpeedAverage; // naming relative to axes in produced picture
			var goMinusH = (byte)(0 < h ? 1 : 0);// <=> 0<x when slice is z, 0<y when slice is x etc.
			switch (sliceDimension)
			{
				case 0: // x
					upDownParticleSpeedAverage = (zData[x, y, z] + zData[x, y, z - goMinusH])/2.0F; //x
					break;
				case 1: // y
					upDownParticleSpeedAverage = (horiz[x, y, z] + horiz[x - goMinusH, y, z])/2.0F; //z
					break;
				default: // z
					upDownParticleSpeedAverage = (vert[x, y, z] + vert[x, y - goMinusH, z])/2.0F; //z
					break;
			}

			// TODO remove:
			const float impedance = 1.2041F * 343;

			float scaledSpeed = impedance * (GMath.Sqrt(
				leftRightParticleSpeedAverage * leftRightParticleSpeedAverage +
				upDownParticleSpeedAverage * upDownParticleSpeedAverage)); // ca 0..1

			float scaledPressure = p[x, y, z] * scalePressure;
			scaledPressure *= scaleBoth;
			scaledSpeed *= scaleBoth;

			float sat = 0F;
			float val = 0F;
			float hue = 0F;

			val = scaledSpeed;
			sat = scaledPressure * 0.5F + 0.5F;

			Clip(ref val);
			val = val * 0.6F + 0.4F; // darkest should be gray (i.e. 0.4), not black (0.0)
			Clip(ref sat);
			sat = sat * 0.7F + 0.2F; // saturation must not be entirely full (i.e. 1.0)
			// or else value aka brightness becomes difficult to see
			hue += 0.1F * sat; // show the saturation value a little bit as color change as well
			Clip(ref hue);
			hue *= 2.0F / 3.0F; // map to the rainbow part of hue range

			if (mat[x, y, z] != 0) // if wall
			{ // set white
				sat = 0F;
				val = 1F;
			}

			byte red;
			byte green;
			byte blue;
			SetHSV(hue, sat, val,
			    out red,
			    out green,
				out blue);

			//if ((mat[x, y, z] & 1) != 0)
			//    red = 255;
			//if ((mat[x, y, z] & 2) != 0)
			//    blue = 255;
			//if ((mat[x, y, z] & 4) != 0)
			//    green = 255;

			dest[w, h, 0] = blue;
			dest[w, h, 1] = green;
			dest[w, h, 2] = red;
			dest[w, h, 3] = 255; // alpha
		}

	}
}



//CudaGPU cudaGPU = (CudaGPU)gpgpu;
//GPGPU castTest = (GPGPU)cudaGPU;
//GASS.CUDA.CUDA cuda = (GASS.CUDA.CUDA)(cudaGPU.CudaDotNet);

//// just an example:
//// syntax for allocating 'pinned' memory:
//_gpu.HostAllocate<int>(3, 4, 5);