﻿#define A

using System;
using Cudafy;
using Cudafy.Host;
using OlaCudafy1.Material;

namespace OlaCudafy1.FDTD
{
	public class SimpleCalc : ICalc
	{
		internal SimpleCalc(CSimulationSpace simSpace, GPGPU gpu)
		{
			_gpu = gpu;
			_simulationSpace = simSpace;
			_thisModule = SuperCudafier.Cudafy(Types);
		}

		private static readonly Type[] Types = new[]
		                                            {
		                                                typeof (KindExtent),
		                                                typeof (Point),
		                                                typeof (GridSettings),
														typeof (Source),
														typeof (MicrophoneStruct),
		                                                typeof (SimpleCalc)
		                                            };

		private readonly CSimulationSpace _simulationSpace;
		private readonly GPGPU _gpu;
		private readonly CudafyModule _thisModule;

		/// <summary>
		/// Lets you specify your own device data to operate on,
		/// as well as your own custom blocks and threads extents,
		/// and your own grid settings
		/// 
		/// Will not interefere with this' CSimulationSpace object reference
		/// </summary>
		public void ManualLaunch(
			dim3 blocks,
			dim3 threads,
			float[,,] devP,
			float[,,] devHoriz,
			float[,,] devVert,
			float[,,] devZSpeeds,
			KindExtent kindE,
			GridSettings[] devGridSettingsArray,
			Source[] devSources,
			MicrophoneStruct[] devMicrophones,
			float[,] devRecordings, 
			byte[,,] devMaterialData,
			byte[,,] devAdjacentSurfaces,
			float[] devAbsorption,
			int generation
			)
		{
			if (_thisModule == null) return;

			_gpu.LoadModule(_thisModule);
#if A
			_gpu.Launch(blocks, threads).SimpleCalcKernel(
				devP,
				devHoriz,
				devVert,
				devZSpeeds,
				kindE,
				devGridSettingsArray,
				0,
				devSources,
				devMicrophones,
				devRecordings,
				devMaterialData,
				devAdjacentSurfaces,
				devAbsorption,
				generation
				);
#else // arguments are rearranged thus does not work lol. why?
			_gpu.Launch(blocks, threads).SimpleCalcKernel(
				devGridSettingsArray,
				devP,
				devHoriz,
				devVert,
				devZSpeeds,
				kindE
				);
#endif
		}

		public void Execute()
		{
			//_gpu.StartTimer();

			var blocks = new dim3(
				Toolbox.RoundUpDiv(_simulationSpace.KindE.X, CudaSettings.BlockDim),
				Toolbox.RoundUpDiv(_simulationSpace.KindE.Y, CudaSettings.BlockDim));
			var threads = new dim3(CudaSettings.BlockDim, CudaSettings.BlockDim);

			ManualLaunch(
				blocks,
				threads,
				_simulationSpace.TheSoundField.DevP,
				_simulationSpace.TheSoundField.DevHorizontalSpeeds,
				_simulationSpace.TheSoundField.DevVerticalSpeeds,
				_simulationSpace.TheSoundField.DevZSpeeds,
				_simulationSpace.KindE,
				_simulationSpace.DevTheGridSettingsArray,
				_simulationSpace.DevTheSources,
				_simulationSpace.DevTheMicrophoneStructs,
				_simulationSpace.DevRecordings,
				_simulationSpace.TheWalls.DevMaterialData,
				_simulationSpace.TheWalls.DevAdjacentSurfaces,
				_simulationSpace.DevAbsorption,
				_simulationSpace.NextGeneration()
				);

			//_elapsedTimeInMilliseconds = _gpu.StopTimer(); // this is pretty slow
		}


		/// <summary>
		/// Calculates one step of Vikholt's reimplementation 
		/// and extension of the Sakamoto laboratory's stagged-grid FDTD algorithm
		/// --- Stripped/Simple Version
		/// </summary>
		[Cudafy]
#if A
		public static void SimpleCalcKernel(GThread thread,
		    float[, ,] p,
		    float[, ,] horiz,
		    float[, ,] vert,
		    float[, ,] zSpeeds,
		    KindExtent ex,
		    GridSettings[] devGridAr,
			byte emulatingPart, // 0 for not emulating
			Source[] srcs,
			MicrophoneStruct[] mics,
			float[,] recordings,
			byte[,,] mat,
			byte[,,] adj,
			float[] abs, // TODO connect with aa and bb
			int generation
		    )
#else
		public static void SimpleCalcKernel(GThread thread,
			GridSettings[] devGridAr,
			float[, ,] p,
			float[, ,] horiz,
			float[, ,] vert,
			float[, ,] zSpeeds,
			KindExtent ex
			)
#endif
		{
			GridSettings grid = devGridAr[0];

			// translate from thread index and block index to pixel position
			int x = thread.threadIdx.x + thread.blockIdx.x*thread.blockDim.x;
			int y = thread.threadIdx.y + thread.blockIdx.y*thread.blockDim.y;

			if (x >= ex.X || y >= ex.Y)
				return;

			const byte W = 0x7;
			int z;

			const float alpha = 0.1f; // see Hill and Hawksford - Viz. and anal. for LF prop.
			float ZN = grid.R0 * grid.C0 * (1.0f + GMath.Sqrt(1.0f - alpha)) / (1.0f - GMath.Sqrt(1.0f - alpha));
			float Rx = 1f / grid.CU;
			float aa = (Rx - ZN) / (Rx + ZN);
			float bb = 2.0f / (Rx + ZN);

			if (emulatingPart == 0 || emulatingPart == 1)
			{
				// =========== v, u, w ============
				for (z = 0; z < ex.Z; ++z)
				{
					var isWallHere = (byte)(mat[x, y, z] & W);
					if (isWallHere == 0 // there's NO wall at the source (to the left) of this particle speed
						|| (x == 0 || y == 0 || (z == 0)) // disregard material at top borders
						)
					{
						if (x < ex.X - 1 && (mat[x + 1, y, z] & W) == 0)
							horiz[x, y, z] += grid.CU*(
							                          	((x > 0 ? 1F : 0F)
							                          	 *p[x, y, z]
							                          	 - p[x + 1, y, z])
							                          );
						else if (x == ex.X - 1)
							horiz[x, y, z] += grid.CU*p[x, y, z];
						else if (x < ex.X - 1)
							horiz[x, y, z] = aa * horiz[x, y, z] + bb * p[x, y, z];
						if (y < ex.Y - 1 && (mat[x, y + 1, z] & W) == 0)
							vert[x, y, z] += grid.CU*(
							                         	((y > 0) ? 1F : 0F)*p[x, y, z]
							                         	- p[x, y + 1, z]
							                         );
						else if (y == ex.Y - 1)
							vert[x, y, z] += grid.CU*p[x, y, z];
						else if (y < ex.Y - 1)
							vert[x, y, z] = aa * vert[x, y, z] + bb * p[x, y, z];
						if (z < ex.Z - 1 && (mat[x, y, z + 1] & W) == 0)
							zSpeeds[x, y, z] += grid.CU*(
							                            	(z > 0 ? 1F : 0F)
							                            	*p[x, y, z]
							                            	- p[x, y, z + 1]);
						else if (z == ex.Z - 1)
							zSpeeds[x, y, z] += grid.CU*p[x, y, z];
						else if (z < ex.Z - 1)
							zSpeeds[x, y, z] = aa * zSpeeds[x, y, z] + bb * p[x, y, z];
						
					} else if (x < ex.X-1 && y<ex.Y-1 && z<ex.Z-1) {
						horiz[x, y, z] = aa * horiz[x, y, z] - bb * p[x + 1, y, z];
						vert[x, y, z] = aa * vert[x, y, z] - bb * p[x, y+1, z];
						zSpeeds[x, y, z] = aa * zSpeeds[x, y, z] - bb * p[x, y, z+1];
						//// WALL_HERE, absorbing wall:
						//const float alpha = 1F; //0.2;
						//float ZN = grid.R0 * grid.C0 *
						//           (1F + GMath.Sqrt(1F - alpha)) / (1F - GMath.Sqrt(1F - alpha));
					}
				}
			}

			if (emulatingPart == 0)
				thread.SyncThreads();

			if (emulatingPart == 0 || emulatingPart == 2)
			{
				// =========== p ============
				for (z = 0; z < ex.Z; ++z)
				{
					if ((mat[x, y, z] & W) == 0)
					{
						if (x > 0 && y > 0 && z > 0)
						{
							p[x, y, z] += grid.CP * (horiz[x - 1, y, z] - horiz[x, y, z]);
							p[x, y, z] += grid.CP * (vert[x, y - 1, z] - vert[x, y, z]);
							p[x, y, z] += grid.CP * (zSpeeds[x, y, z - 1] - zSpeeds[x, y, z]);
						}
					}
					else
					{
						// vanlig:
						p[x, y, z] = 0F;
						// impedance wall test:
						//float S = grid.DX;
						//const float k = .8F;
						//float C = grid.R0*grid.C0*S; // <-
						//const float m = 0.00F;
						//float denominator = m + C*grid.DT + k*grid.DT*grid.DT;
						//float konst = 2*C*grid.DT + k*grid.DT*grid.DT;
						//float dxdt = horiz[x, y, z];
						//p[x, y, z] += dxdt;
						//dxdt = (dxdt*m - konst*p[x, y, z] - p[x + 1, y, z]*S*grid.DT)/denominator;
						//horiz[x, y, z] = dxdt;
					}

					// todo implement different source types
					foreach (var source in srcs)
					{
						// impulse
						if (source.InvokeAtGeneration == generation && source.Type == 0)
						{
							const float spikeBigness = 0.02f;// 0.002f;
							p[x, y, z] = source.PressureAmplitude
							             *GMath.Exp(-(
							                         	(x - source.X - 0.5f)*grid.DX*(x - source.X - 0.5f)*grid.DX +
							                         	(y - source.Y - 0.5f)*grid.DX*(y - source.Y - 0.5f)*grid.DX +
							                         	(z - source.Z - 0.5f)*grid.DX*(z - source.Z - 0.5f)*grid.DX
							                         )/spikeBigness);
						}
							// sweep
						else if (source.Type == 1)
						{
							if (x != source.X
								|| y != source.Y
								|| z != source.Z)
								continue;
							float time = generation*grid.DT;
							const float startFrequency = 15f;
							//float sweepFrequency = startFrequency + grid.GridNyquistFreq * time;
							float sweepFrequency = startFrequency + grid.GridMaxFreq*time;
							float angularSweepFrequency = 2*GMath.PI*sweepFrequency;
							p[x, y, z] = source.PressureAmplitude*GMath.Sin(angularSweepFrequency*time);
							//float angularFrequency = 2 * GMath.PI * source.Frequency;
							//p[x, y, z] = source.PressureAmplitude * GMath.Sin(angularFrequency * time);
						}
					}
					for (int index = 0; index < mics.Length; index++)
					{
						var mic = mics[index];
						if (x != mic.X ||
						    y != mic.Y ||
						    z != mic.Z)
							continue;
						recordings[index, generation] = p[x, y, z];
					}
				}
			}

		}


		private float _elapsedTimeInMilliseconds;

		public float ElapsedTimeInMilliseconds
		{
			get { return _elapsedTimeInMilliseconds; }
		}

	}
}
