﻿using System;
using Cudafy;
using Cudafy.Host;

namespace OlaCudafy1.FDTD
{
	internal class Wall<T> : IDisposable // aka Walls, or CellConfig, or Cells
	{
		private readonly GPGPU _gpu;

		/// <summary>
		/// Stores the material type of each voxel.
		/// Will contain an index which will be used for a lookup
		/// in an impedance database in the future.
		/// </summary>
		private readonly T[, ,] _materialData;

		private T _airValue;

		/// <summary>
		/// Indexer which sets a material voxel in host memory.
		/// 
		/// Invalidates the GPU copy so that it will be updated
		/// before it is used.
		/// </summary>
		internal T this[int x, int y, int z]
		{
			set {
				_materialData[x, y, z] = value;
				InvalidateDevMaterialData();

				RecalculateAndSetAdjacentSurfacesForNewMaterialAtPoint(x, y, z, value);
				InvalidateDevAdjacentSurfaces();
			}
		}

		internal int Length
		{
			get { return _materialData.Length; }
		}

		private bool IsAir(T value)
		{
			return (value.Equals(_airValue));
		}

		private void RecalculateAndSetAdjacentSurfacesForNewMaterialAtPoint(int x, int y, int z, T value)
		{
			if (IsAir(value))
			{ // clear bit if adjacent to air
				if (x < _kindExtent.X - 1)
					_adjacentSurfaces[x + 1, y, z] &= (0xff ^ 1 << 0);
				if (y < _kindExtent.Y - 1)
					_adjacentSurfaces[x, y + 1, z] &= (0xff ^ 1 << 1);
				if (z < _kindExtent.Z - 1)
					_adjacentSurfaces[x, y, z + 1] &= (0xff ^ 1 << 2);
				if (x > 0)
					_adjacentSurfaces[x - 1, y, z] &= (0xff ^ 1 << 3);
				if (y > 0)
					_adjacentSurfaces[x, y - 1, z] &= (0xff ^ 1 << 4);
				if (z > 0)
					_adjacentSurfaces[x, y, z - 1] &= (0xff ^ 1 << 5);
			} else { // set bit otherwise
				if (x < _kindExtent.X - 1)
					_adjacentSurfaces[x + 1, y, z] |= 1 << 0;
				if (y < _kindExtent.Y - 1)
					_adjacentSurfaces[x, y + 1, z] |= 1 << 1;
				if (z < _kindExtent.Z - 1)
					_adjacentSurfaces[x, y, z + 1] |= 1 << 2;
				if (x > 0)
					_adjacentSurfaces[x - 1, y, z] |= 1 << 3;
				if (y > 0)
					_adjacentSurfaces[x, y - 1, z] |= 1 << 4;
				if (z > 0)
					_adjacentSurfaces[x, y, z - 1] |= 1 << 5;
			}
		}

		private readonly T[, ,] _devMaterialData;

		/// <summary>
		/// Gets the pointer to the material data on the device aka GPU.
		/// Will be updated to reflect the material data on the host.
		/// </summary>
		internal T[,,] DevMaterialData
		{
			get
			{
				if (DevMaterialDataNeedsUpdate)
					UpdateDevMaterial();
				return _devMaterialData;
			}
		}

		/// <summary>
		/// Marks the surfaces of each wall voxel
		/// which are NOT adjacent to air voxels.
		/// The impedance of these surfaces must be simulated in the
		/// GPU kernel function, while the surfaces which are adjacent to
		/// other wall voxels, must not be simulated.
		/// 
		/// The byte at [x,y,z] will be interpreted bitwise,
		/// to mark the 0-6 non-free surfaces of the wall voxel
		/// at (x,y,z).
		/// 
		/// Here, the 0th bit i.e. byte & 1&lt;&lt;0
		/// is true if the positive x side lies adjacent
		/// to a material wall and false if it lies adjacent 
		/// to (an) air (wall element).
		/// 
		/// The 1st bit i.e. byte & 1&lt;&lt;1 refers to
		/// the positive y side, and the 2nd bit (byte & 4)
		/// to the positive z side.
		/// 
		/// Continuing in the same fashion,
		///  the 3rd, 4th and 5th bit refers to the
		/// negative x, y and z sides respectively.
		/// 
		/// This variable/field must be updated to reflect
		/// material data whenever that changes.
		/// </summary>
		private byte[,,] _adjacentSurfaces;
		internal byte[, ,] AdjacentSurfaces { get { return _adjacentSurfaces; } }
		private byte[,,] _devAdjacentSurfaces;

		internal byte[, ,] DevAdjacentSurfaces
		{
			get
			{
				if (DevAdjacentSurfacesNeedsUpdate)
					UpdateDevAdjacentSurfaces();
				return _devAdjacentSurfaces;
			}
		}

		public Wall(KindExtent kindExtent, GPGPU gpu, T airValue)
		{
			if (!airValue.Equals((byte)0))
				throw new NotImplementedException("The current implementation does not support the air value " +
				                                  "being anything else than 0, as it uses default array " +
				                                  "initialization.");
			_gpu = gpu;
			_airValue = airValue;
			_kindExtent = kindExtent;
			// data starts out 0 i.e. as air
			Allocate<T>(_gpu, out _materialData, out _devMaterialData, _kindExtent);
			Allocate<byte>(_gpu, out _adjacentSurfaces, out _devAdjacentSurfaces, _kindExtent);
			DevMaterialDataUpdated();
		}

		private void FreeGPUMemory()
		{
			if (_gpu.IsOnGPU(_devMaterialData))
				_gpu.Free(_devMaterialData);
			if (_gpu.IsOnGPU(_devAdjacentSurfaces))
				_gpu.Free(_devAdjacentSurfaces);
		}

		~Wall()
		{
			Dispose();
		}

		public void Dispose()
		{
			FreeGPUMemory();
		}

		private static void Allocate<TU>(
			GPGPU gpu,
			out TU[,,] data,
			out TU[,,] devData,
			KindExtent kindExtent)
		{
			data = new TU[
				kindExtent.X,
				kindExtent.Y,
				kindExtent.Z];
			devData = gpu.Allocate(data);
		}

		private void UpdateDevMaterial()
		{
			if (!_gpu.IsOnGPU(_devMaterialData))
				throw new Exception("Data was not on GPU O__O.");
			_gpu.CopyToDevice(hostArray: _materialData, devArray: _devMaterialData);
			DevMaterialDataUpdated();
		}

		private void UpdateDevAdjacentSurfaces()
		{
			if (!_gpu.IsOnGPU(_devAdjacentSurfaces))
				throw new Exception("Data was not on GPU O__O.");
			_gpu.CopyToDevice(hostArray: _adjacentSurfaces, devArray: _devAdjacentSurfaces);
			DevAdjacentSurfacesUpdated();
		}

		private bool _isInvalidatedDevMaterialData;
		private void DevMaterialDataUpdated()
		{
			_isInvalidatedDevMaterialData = false;
		}
		private void InvalidateDevMaterialData()
		{
			_isInvalidatedDevMaterialData = true;
		}
		private bool DevMaterialDataNeedsUpdate
		{
			get { return _isInvalidatedDevMaterialData; }
		}

		private bool _isInvalidatedDevAdjacentSurfaces;
		private KindExtent _kindExtent;
		private void DevAdjacentSurfacesUpdated()
		{
			_isInvalidatedDevAdjacentSurfaces = false;
		}
		private void InvalidateDevAdjacentSurfaces()
		{
			_isInvalidatedDevAdjacentSurfaces = true;
		}
		private bool DevAdjacentSurfacesNeedsUpdate
		{
			get { return _isInvalidatedDevAdjacentSurfaces; }
		}
	}
}