﻿using System;
using Cudafy;
using GASS.CUDA;
using OlaCudafy1.FDTD;

namespace OlaCudafy1.Material
{
	internal class Extent
	{
		private readonly Range[] _floatRanges;

		private KindExtent _kindE;

		private const int Three = 3; // x, y and z

		public Extent(int x, int y, int z)
		{
			KindE = new KindExtent {X = x, Y = y, Z = z};
			_floatRanges = new Range[Three];
		}

		public Extent(int[] axes)
		{
			if (axes.Length != 3)
				throw new ExtentDimensionException("Can only make three-dimensional extents, " +
				                                   "i.e. from axes of length 3. Supplied axes " +
				                                   "had length " + axes.Length);
			KindE = new KindExtent {X = axes[0], Y = axes[1], Z = axes[2]};
			_floatRanges = new Range[axes.Length];
		}

		/// <summary>
		/// Returns the length of the longest dimension (whichever is longest of x, y and z)
		/// </summary>
		/// <returns></returns>
		public int GetMax()
		{
			return Math.Max(Math.Max(X, Y), Z);
		}

		public Extent(KindExtent kindExtent)
		{
			_kindE = kindExtent;
		}

	

		public int Rank
		{
			get { return Three; }
		}

		internal Range[] FloatRanges
		{
			get { return _floatRanges; }
		}

		public int X
		{
			get { return KindE.X; }
		}

		public int Y
		{
			get { return KindE.Y; }
		}

		public int Z
		{
			get { return KindE.Z; }
		}

		public KindExtent KindE
		{
			get { return _kindE; }
			set { _kindE = value; }
		}

		/// <summary>
		/// Produces a cuda extent. A cuda extent has, for the x direction,
		///  not the number of elements but the number of bytes,
		///  i.e. the number of elements times the byte-count of
		///  the datatype (as returned by sizeof).
		/// For the y and z direction, it's the same as  what you would expect
		/// in a regular extent, i.e. the number of elements.
		/// </summary>
		/// <returns>Returns the new cuda extent.</returns>
		public cudaExtent ToCudaExtent()
		{
			return new cudaExtent {height = Y, depth = Z, width = X*sizeof (float)}; // todo
		}


		private void SetFloatRange(int dimension, float min, float max)
		{
			_floatRanges[dimension] = new Range(min, max);
		}

		/// <summary>
		/// Specifies what (float) range is in use on the axis along 'dimension'
		/// </summary>
		/// <param name="dimension">The axis whose range to set; x, or y, etc.</param>
		/// <param name="min">The coordinate of the left-most voxel</param>
		/// <param name="max">The coordinate of the right-most voxel</param>
		internal void SetFloatRange(char dimension, float min, float max)
		{
			switch (dimension)
			{
				case 'x':
					SetFloatRange(0, min, max);
					break;
				case 'y':
					SetFloatRange(1, min, max);
					break;
				case 'z':
					SetFloatRange(2, min, max);
					break;
			}
		}

		[CudafyIgnore]
		internal float GetStepSize(int dimension)
		{
			int ting = 0;
			switch(dimension)
			{
				case 0:
					ting = (X - 1); // number of cell intervals
					break;
				case 1:
					ting = (Y - 1);
					break;
				case 2:
					ting = (Z - 1);
					break;																		   

			}
			// average of differences between successive cells
			return _floatRanges[dimension].Span / ting;
		}

		public int this[SliceDimension sliceDimension]
		{
			get
			{
				switch (sliceDimension)
				{
					case SliceDimension.X:
						return X;
					case SliceDimension.Y:
						return Y;
					case SliceDimension.Z:
						return Z;
					default:
						return 0;
				}  /* return the specified index here */
			}
		}

		public int this[int index]
		{
			get
			{
				switch (index)
				{
					case 0:
						return X;
					case 1:
						return Y;
					case 2:
						return Z;
					default:
						return 0;
				}  /* return the specified index here */
			}
		}
	}
}