﻿using System.Collections.Generic;
using System.Linq;
using System.Xml.Linq;

namespace OlaCudafy1.Material
{
	internal class MaterialModel
	{
		// convenience thing
		private static readonly string[] XYZ = new[] {"x", "y", "z"};

		private Extent _extent;
		private int _cellCount;
		private Cell[] _cells;
		private readonly Dictionary<string,int> _materialNames = new Dictionary<string,int>();

		public Dictionary<string, int> MaterialNames
		{
			get { return _materialNames; }
		}

		public IList<string> ProduceMaterialsArray()
		{
			var ar = new string[_materialNames.Count];
			foreach (var materialName in _materialNames)
			{
				ar[materialName.Value] = materialName.Key;
			}
			return ar;
		}

		public Extent Extent
		{
			get { return _extent; }
		}

		internal Cell[] Cells
		{
			get { return _cells; }
			set { _cells = value; }
		}



		/// <summary>
		/// Guesses the number of voxel positions in each direction.
		/// (Assumes that each voxel plane normal to a dimension d
		/// is filled at least at once by a voxel,
		/// for d in each of the three dimensions x, y and z.)
		/// </summary>
		/// <param name="docRoot"></param>
		/// <returns></returns>
		private static Extent GuessExtent(XContainer docRoot)
		{
			// distinct x y z
			var axes = new int[XYZ.Length];
			int axesIndex = 0;
			foreach (string c in XYZ)
			{
				IEnumerable<float> d = (docRoot.Descendants(c).Select(e => (float) e)).Distinct();
				axes[axesIndex++] = d.Count();
			}
			return new Extent(axes);
		}

		/// <summary>
		/// Extracts the different material names from the XML document
		/// </summary>
		/// <param name="docRoot"></param>
		/// <returns></returns>
		private static IEnumerable<string> GetMaterialNames(XContainer docRoot)
		{
			// different types of material
			IEnumerable<string> uniqmat = (
				docRoot.Descendants("material").Select(e => e.Value)).Distinct();
			return uniqmat.ToArray();
		}

		/// <summary>
		/// Creates an empty material model (no walls) of the specified extent aka size
		/// </summary>
		/// <param name="ex"></param>
		public MaterialModel(Extent ex)
		{
			_cellCount = 0;
			_cells = null;
			_materialNames = null; 
			FromExtentOpenSpace(ex);
		}
		
		/// <summary>
		/// Creates a material model with walls imported from the Westerdiep compliant
		/// XML file pointed to by the filename
		/// </summary>
		/// <param name="filename"></param>
		public MaterialModel(string filename)
		{
			LoadAndProcess(filename);
		}

		/// <summary>
		/// Creates a material model from the supplied Westerdiep compliant
		/// XML document
		/// </summary>
		/// <param name="doc"></param>
		public MaterialModel(XDocument doc)
		{
			LoadAndProcessDoc(doc);
		}

		/// <summary>
		/// Creates an empty material model (no walls) of the specified extent aka size
		/// </summary>
		/// <param name="extent"></param>
		private void FromExtentOpenSpace(Extent extent)
		{
			_extent = extent;
		}

		private XDocument LoadXDocumentFromFile(string filename)
		{

			return XDocument.Load(filename);
		}

		private void LoadAndProcess(string filename)
		{
			var doc = LoadXDocumentFromFile(filename);

			LoadAndProcessDoc(doc);
		}

		private void LoadAndProcessDoc(XDocument doc)
		{
			var docRoot = doc.Root;
			if (docRoot == null) return;

			_extent = GuessExtent(doc.Root);
			int materialIndex = 0;
			foreach (string materialName in GetMaterialNames(doc.Root))
			{
				_materialNames.Add(materialName, materialIndex);
				++materialIndex;
			}
			
			// total cell (aka cube) count
			_cellCount = doc.Descendants("position").Count();

			// coordinate ranges and step size for each dimension x, y and z
			foreach (string dim in XYZ)
			{
				float min = doc.Descendants(dim).Min(e => (float) e);
				float max = doc.Descendants(dim).Max(e => (float) e);
				_extent.SetFloatRange(dim[0], min, max);
			}

			// process all cells
			_cells = new Cell[_cellCount];
			int cellIndex = 0;
			IEnumerable<XElement> xmlCells = docRoot.Elements("position");
			foreach (XElement xmlCell in xmlCells)
			{
				var cell = new Cell();

				XElement xElement = xmlCell.Element("material");
				if (xElement != null)
				{

					if (!_materialNames.ContainsKey(xElement.Value)) // that's strange
						_materialNames[xElement.Value] = _materialNames.Count + 1;
					cell.MaterialIndex = _materialNames[xElement.Value];
				}
				
				var position = new Point();
				for (int dim = 0; dim < XYZ.Length; dim++)
				{
					string dimName = XYZ[dim];

					// subtract the lowest coordinate in the /dim/ dimension
					// so that the cells are "left"-adjusted to (0,0,0) and
					// lie inside the (+,+,+)-/positive octant
					float cellPositionAdjusted = (float) xmlCell.Element(dimName)
					                             - _extent.FloatRanges[dim].Min;

					// divide by cell spacing and round off to integer voxel index
					var cellPositionInt = (int) (cellPositionAdjusted/_extent.GetStepSize(dim) + 0.5F);
					position[dim] = cellPositionInt;
				}
				cell.Position = position;
				_cells[cellIndex] = cell; // add processed cell to collection
				++cellIndex;
			}
		}
	}
}