﻿using System;
using System.ComponentModel;
using Cudafy;

namespace OlaCudafy1
{
	/// <summary>
	/// The RindelImpedanceEstimation class
	/// provides logic for estimating the absorber type and 
	/// the three parameters of a simple material impedance model
	/// of the absorber, given the absorption coefficients of
	/// the material at 125, 250, 500 and 1000 Hz.
	/// 
	/// This can be used in for example the FDTD simulation method
	/// to model the impedance of a material for which only
	/// the absorption coefficients are known. The absorption 
	/// coefficients for many materials are easily available
	/// through large databases.
	/// 
	/// This implementation is based on the method described
	/// by J. H. Rindel in 'An impedance model for estimating
	///  the complex pressure reflection factor', presented at
	/// Forum Acusticum 2011 in Aalborg, Denmark.
	/// </summary>
	internal class RindelImpedanceEstimation
	{
		#region AbsorberType enum

		public enum AbsorberType
		{
			Porous, // section 3.4 in the article
			Membrane, // section 3.5 in the article
			Resonator, // section 3.6 in the article
		}

		#endregion

		private readonly AbsorberType _absorberType;
		private readonly float[] _absorptionCoefficients;
		private readonly float _acousticImpedance;
		private readonly BandCoefficients _bandCoefficients;
		private readonly FrequencyBand _whatBandToUse;
		private BandCoefficients _imaginaryTerms;

		/// <summary>
		/// </summary>
		/// <param name="absorptionCoefficients">
		/// The statistic absorption coefficients alpha_stat of the material.</param>
		/// <param name="acousticImpedance">
		/// The acoustic impedance to use in the calculations.</param>
		internal RindelImpedanceEstimation(float[] absorptionCoefficients, float acousticImpedance)
		{
			_absorptionCoefficients = absorptionCoefficients;
			_absorberType = GuessAbsorberType(_absorptionCoefficients);
			_bandCoefficients = new BandCoefficients(_absorptionCoefficients);
			_acousticImpedance = acousticImpedance;
			if (_absorberType == AbsorberType.Resonator)
				_whatBandToUse = _bandCoefficients.BandOfMaximum;
			else
				_whatBandToUse = WhatBandToUse(_absorberType);
			EstimatedResistanceTerm = EstimateResistanceTerm();
			EstimateImaginaryTerms();
			EstimatedImaginaryTermsArray = _imaginaryTerms.ToArray();
		}

		internal float EstimatedResistanceTerm { get; private set; }
		internal float[] EstimatedImaginaryTermsArray { get; private set; }

		private float FrequencyBandFrequencyValue(FrequencyBand band)
		{
			switch (band)
			{
				case FrequencyBand.Band125Hz:
					return 125;
				case FrequencyBand.Band250Hz:
					return 250;
				case FrequencyBand.Band500Hz:
					return 500;
				case FrequencyBand.Band1000Hz:
					return 1000;
				default:
					throw new InvalidEnumArgumentException();
			}
		}

		/// <summary>
		/// Guesses the absorber type according to the method described by
		/// J. H. Rindel, on the basis of the supplied absorption coefficients.
		/// The supplied float values should be the absorption coefficient
		/// (alpha_stat) values for the four frequency bands
		/// 125, 250, 500 and 1000 Hz, upon which the absorber type will be
		/// guessed.
		/// </summary>
		/// <returns></returns>
		private static AbsorberType GuessAbsorberType(
			BandCoefficients bandCoefficients)
		{
			float alpha0125HzBand = bandCoefficients.Alpha125HzBand;
			float alpha0250HzBand = bandCoefficients.Alpha250HzBand;
			float alpha0500HzBand = bandCoefficients.Alpha500HzBand;
			float alpha1000HzBand = bandCoefficients.Alpha1000HzBand;
			if (alpha0125HzBand <= alpha0250HzBand &&
			    alpha0250HzBand <= alpha0500HzBand &&
			    alpha0500HzBand <= alpha1000HzBand) // monotonically increasing absorption
				return AbsorberType.Porous;
			if (alpha0125HzBand >= alpha0250HzBand &&
			    alpha0250HzBand >= alpha0500HzBand) // monotonically decreasing
				return AbsorberType.Membrane;
			return AbsorberType.Resonator; // otherwise
		}

		/// <summary>
		/// Guesses the absorber type according to the method described by
		/// J. H. Rindel, on the basis of the supplied absorption coefficients.
		/// The supplied float array should be of length 5 but may be longer,
		/// and will be interpreted as follows:
		///	  1st element: 63 Hz: ignored;
		///   2nd-5th element: 125, 250, 500 and 1000 Hz: used for estimation;
		///   any susequent elements if present: ignored;
		/// </summary>
		/// <param name="absorptionCoefficients"></param>
		/// <returns></returns>
		public static AbsorberType GuessAbsorberType(float[] absorptionCoefficients)
		{
			var bandCoefficients = new BandCoefficients(absorptionCoefficients);
			return GuessAbsorberType(bandCoefficients);
		}

		/// <summary>
		/// Estimates the imaginary term X(f) for each of the bands (f=)
		/// 125, 250, 500 and 1000 Hz and places them in a float array looking like this: 
		/// [0, (125Hz-term), (250Hz-term), (500Hz-term), (1000Hz-term)].
		/// Estimation is done for an absorber having the supplied coefficients.
		/// Estimation is done according Rindel's method, section 3.2, equation (9).
		/// 
		/// X(f) can be positive or negative, but only the positive values are returned.
		/// The other values can be found by taking -X(f)
		/// </summary>
		/// <returns></returns>
		private void EstimateImaginaryTerms()
		{
			float R = EstimatedResistanceTerm;
			_imaginaryTerms = new BandCoefficients();
			foreach (FrequencyBand band in Enum.GetValues(typeof (FrequencyBand)))
			{
				float zfStar = ZfStar.Get(band, _acousticImpedance);
				float nominator = 4f*R*zfStar;
				float denom = _bandCoefficients[band];
				float term = (R + zfStar);
				var x = (float) Math.Sqrt(nominator/denom - term*term);
				_imaginaryTerms[band] = x;
			}
			// TODO
			// In addition, imaginaryTerms[0] should contain data for the 63Hz band,
			// (and possibly imaginaryTerms[5] for the 2000Hz band)
			// estimated by using Rindel's equation (3) in section 2.1, using:
			// omega=2*PI*63Hz, and with 'd' and 'm' as first estimated according to the
			// scattered instructions in the article which differ depending on absorber type.
		}

		/// <summary>
		/// Estimates 'd' according to equation (13)
		/// </summary>
		/// <returns></returns>
		internal float EstimateD(float speedOfSound)
		{
			switch (_absorberType)
			{
				case AbsorberType.Porous:
					float nom = _acousticImpedance*speedOfSound;
					float freq = FrequencyBandFrequencyValue(FrequencyBand.Band125Hz);
					float den = freq*2f*
					            (float) Math.PI*Math.Abs(_imaginaryTerms[FrequencyBand.Band125Hz]);

					float nom250 = _acousticImpedance*speedOfSound;
					float freq250 = FrequencyBandFrequencyValue(FrequencyBand.Band250Hz);
					float den250 = freq250*2f*
					               (float) Math.PI*Math.Abs(_imaginaryTerms[FrequencyBand.Band250Hz]);

					return (nom/den + nom250/den250)*.5f; // average
				case AbsorberType.Membrane:
					return float.PositiveInfinity;
				case AbsorberType.Resonator: // equation (16)
					//// TODO consolidate with EstimateM
					float nomres = _acousticImpedance* speedOfSound;
					 // dupe from estimate m -> res
					float omega1 = GMath.PI*2f*FrequencyBandFrequencyValue(FrequencyBand.Band125Hz);
					 // dupe from estimate m -> res
					float nom1 = Math.Abs(_imaginaryTerms[FrequencyBand.Band125Hz])*omega1;
					float denres = omega1*omega1*EstimateM() + nom1;
					return nomres/denres;
				default:
					throw new NotImplementedException();
			}
		}

		internal float EstimateM()
		{
			switch (_absorberType)
			{
				case AbsorberType.Porous:
					return 0f;

				case AbsorberType.Membrane:
					float omega500 = 2f*GMath.PI*FrequencyBandFrequencyValue(FrequencyBand.Band500Hz);
					float nom500 = Math.Abs(_imaginaryTerms[FrequencyBand.Band500Hz]);
					float den500 = omega500;

					float omega250 = 2f*GMath.PI*FrequencyBandFrequencyValue(FrequencyBand.Band250Hz);
					float nom250 = Math.Abs(_imaginaryTerms[FrequencyBand.Band250Hz]);
					float den250 = omega250;

					return (nom500/den500 + nom250/den250)*.5f; // average

				case AbsorberType.Resonator: // refer to equation (15)
					float omega1 = GMath.PI*2f*FrequencyBandFrequencyValue(FrequencyBand.Band125Hz);
					float omega2 = GMath.PI*2f*FrequencyBandFrequencyValue(FrequencyBand.Band1000Hz);
					float nom1 = Math.Abs(_imaginaryTerms[FrequencyBand.Band125Hz])*omega1;
					float nom2 = Math.Abs(_imaginaryTerms[FrequencyBand.Band1000Hz])*omega2;
					float nom = nom1 + nom2;
					float den = omega2*omega2 - omega1*omega1;
					return nom/den;

				default:
					throw new NotImplementedException();
			}
		}

		/// <summary>
		/// Taken from definitions at Rindel's article section 3.4 and 3.5.
		/// </summary>
		/// <param name="absorberType"></param>
		/// <returns></returns>
		private static FrequencyBand WhatBandToUse(AbsorberType absorberType)
		{
			switch (absorberType)
			{
				case AbsorberType.Porous:
					return FrequencyBand.Band1000Hz;
				case AbsorberType.Membrane:
					return FrequencyBand.Band125Hz;
				default:
					throw new InvalidEnumArgumentException("Illegal AbsorberType enum type passed");
			}
		}

		/// <summary>
		/// Estimates the resistance term R of the simple impedance model for the
		/// absorber with supplied coefficients, according to Rindel's method,
		/// in Rindel's article section 3.3 equation (12).
		/// 
		/// The unit of the returned R is kg per square meter per second.
		/// </summary>
		/// <returns></returns>
		private float EstimateResistanceTerm()
		{
			if (_absorberType == AbsorberType.Porous || _absorberType == AbsorberType.Membrane
				 || _absorberType == AbsorberType.Resonator)
			{
				float g = G(_whatBandToUse, _bandCoefficients[_whatBandToUse], _acousticImpedance);
				var expression = (float) Math.Sqrt((
				                 g*g - ZfStar.GetSquared(_whatBandToUse, _acousticImpedance)));
				ResistanceTermSolutionType resistanceTermSolutionType =
					ResistanceTermSolutionType.Hard;
				if (_absorberType == AbsorberType.Porous)
					resistanceTermSolutionType = _bandCoefficients.Maximum > 0.20
					                             	? ResistanceTermSolutionType.Soft
					                             	: ResistanceTermSolutionType.Hard;
				else if (_absorberType == AbsorberType.Membrane)
					// always hard in case of membrane
					resistanceTermSolutionType = ResistanceTermSolutionType.Hard;
				else if (_absorberType == AbsorberType.Resonator)
					resistanceTermSolutionType = ResistanceTermSolutionType.Soft;
				float R = g +
				          (resistanceTermSolutionType == ResistanceTermSolutionType.Hard
				           	? expression
				           	: -expression);
				return R;
			}
			throw new InvalidEnumArgumentException(
				string.Format("absorberType can only be one of {0}, {1} or {2}, " + "but was {3}.",
				              AbsorberType.Porous.ToString(),
				              AbsorberType.Membrane.ToString(),
				              AbsorberType.Resonator.ToString(),
				              _absorberType.ToString()));
		}

		/// <summary>
		/// Implements the G abbreviate function from section 3.2.
		/// </summary>
		/// <returns></returns>
		private static float G(FrequencyBand omega, float alphaS, float acousticImpedance)
		{
			return ZfStar.Get(omega, acousticImpedance)*(2f/alphaS - 1);
		}

		#region Nested type: ResistanceTermSolutionType

		private enum ResistanceTermSolutionType
		{
			Hard,
			Soft
		}

		#endregion

		#region Nested type: ZfStar

		/// <summary>
		/// Picks out values from the table in section 3.1 in Rindel's article.
		/// </summary>
		private class ZfStar
		{
			/// <summary>
			/// Get the value of Z_f_star, i.e. the equivalent field impedance
			/// relative to the impedance of air, for the given frequency band
			/// </summary>
			/// <returns></returns>
			public static float Get(FrequencyBand band, float acousticImpedance)
			{
				// CAUTION! assumes a value for rho*c
				//const float rhoC = 1.2041f * 343.26f;
				//const float rhoC = 1.2250f * 340.31f;
				//return GetOverRhoC(band)*rhoC;
				return GetOverRhoC(band)*acousticImpedance;
			}

			private static float GetOverRhoC(FrequencyBand band)
			{
				switch (band)
				{
					case FrequencyBand.Band125Hz:
						return 1.04f; // see the article
					case FrequencyBand.Band250Hz:
						return 1.35f;
					case FrequencyBand.Band500Hz:
						return 1.53f;
					case FrequencyBand.Band1000Hz:
						return 1.62f;
					default:
						throw new InvalidEnumArgumentException(
							"Illegal FrequencyBand enum type passed");
				}
			}

			public static float GetSquared(FrequencyBand band, float acousticImpedance)
			{
				return Get(band, acousticImpedance)*Get(band, acousticImpedance);
			}
		}

		#endregion
	}
}