﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Globalization;
using System.Text;
using System.Windows.Forms;
using Cudafy;
using Cudafy.Host;
using GASS.CUDA;
using OlaCudafy1.FDTD;
using OlaCudafy1.Material;
using OlaCudafy1.Properties;
using ZedGraph;

namespace OlaCudafy1
{
	public partial class SakaForm : Form
	{
		private static string _lastFileName = string.Empty;
		private static OpenFileDialog _openXMLFileDialog;
		private readonly Timer _freeMemoryDelayBeforeUpdateFreeMemoryLabelTimer = new Timer();
		private FormAbsorberDatabase _absorberForm;

		private ICalc _calc;
		private DisplaySettings _displaySettingsZY;
		private DisplaySettings _displaySettingsXY;
		private DisplaySettings _displaySettingsXZ;
		private FDTDDrawMachine _drawMachineZY;
		private FDTDDrawMachine _drawMachineXY;
		private FDTDDrawMachine _drawMachineXZ;

		/// <summary>
		/// Redraws the sound volume slice views visible in this form,
		/// by first launching draw machine and then updating picture box
		/// </summary>
		private void FDTDRedraw() // TODO: permanently associate picture boxes and draw machines
		{
			if (_drawMachineXZ != null)
			{
				_drawMachineXZ.Launch();

				pictureBoxZY.Image = new Bitmap(_drawMachineXZ.Bitmap, pictureBoxZY.Size);
				pictureBoxZY.Refresh();
			}

			if (_drawMachineXY != null)
			{
				_drawMachineXY.Launch();
				pictureBoxXY.Image = new Bitmap(_drawMachineXY.Bitmap, pictureBoxXY.Size);
				pictureBoxXY.Refresh();
			}

			if (_drawMachineZY != null)
			{
				_drawMachineZY.Launch();
				pictureBoxXZ.Image = new Bitmap(_drawMachineZY.Bitmap, pictureBoxXZ.Size);
				pictureBoxXZ.Refresh();
			}

		}

		/// <summary>
		/// The GPU currently in use
		/// </summary>
		private GPGPU _gpu;

		private MaterialModel _materialModel;
		private Microphone[] _microphones;

		private CSimulationSpace _simulationSpace;
		private float _simulationSpaceLargestDimension = float.NaN;
		private Source[] _sources;
		private WallAbsorberAssociations _wallAbsorberAssociations;

		private Timer _runSimulationTimer;

		private OlaCufft _olaCufft;

		public SakaForm()
		{
			InitializeComponent();

			_freeMemoryDelayBeforeUpdateFreeMemoryLabelTimer.Tick +=
				_freeMemoryDelayBeforeUpdateFreeMemoryLabelTimer_Tick;

			_freeMemoryDelayBeforeUpdateFreeMemoryLabelTimer.Interval = 1000;

			_runSimulationTimer = new Timer();
			_runSimulationTimer.Tick += runSimulationTimer_Tick;
			_runSimulationTimer.Interval = 10;

			_olaCufft = new OlaCufft(zg1);
		}

		/// <summary>
		/// Gets the Extent of the currently loaded Material Model.
		/// (Assumes that we only have one material model loaded at a time.)
		/// </summary>
		internal Extent Extent
		{
			get { return _materialModel.Extent; }
		}

		/// <summary>
		/// Quits the application
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void ButtonQuitClick(object sender, EventArgs e)
		{
			Application.Exit();
		}

		/// <summary>
		/// Iterates the simulation by one time step.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void ButtonIterateClick(object sender, EventArgs e)
		{
			if (_calc == null) return;

			_calc.Execute();

			//textBoxDebugOutput.Text +=
			//    string.Format("Time elapsed: {0} ms",
			//                  _calc.ElapsedTimeInMilliseconds) + Environment.NewLine;

			Focus();
			Refresh();

			FDTDRedraw();
		}


		private void ButtonPrepareClick(object sender, EventArgs e)
		{
			if (_materialModel == null) return;

			int largestDimensionCellCount = _materialModel.Extent.GetMax();
			if (largestDimensionCellCount == 0)
			{
				textBoxDebugOutput.Text +=
					"Largest dimension cell count should be positive, not 0. " +
					"Probably you should remember to press the make extent button. " +
					"Not continuing." + Environment.NewLine;
				return;
			}
			if (float.IsNaN(_simulationSpaceLargestDimension))
			{
				textBoxDebugOutput.Text +=
					"You must set the largest space dimension first. " +
					"Not preparing." + Environment.NewLine;
				return;
			}

			float dx = _simulationSpaceLargestDimension/largestDimensionCellCount;
			var gridSettings = new GridSettings();
			gridSettings.Setup(dx);

			try
			{
				_gpu = CudafyHost.GetDevice(CudafyModes.Target);
			}
			catch (CUDAException cudaException)
			{
				if (cudaException.CUDAError == CUResult.ErrorOutOfMemory)
				{
					textBoxDebugOutput.Text +=
						"CUDA is out of memory. Try restarting the application." + Environment.NewLine;
					return;
				} // else
				throw;
			}
			_olaCufft.SetGPU(_gpu);

			_sources = new[] {new Source(Extent.X/2, Extent.Y/2, Extent.Z/2, 100F, 0.020f),}; // 100Hz, 60dB
			_microphones = new[] {new Microphone(0, 0, 0)};

			// for odeon-test L-shaped room:
			int sourceX = (int) Math.Round(4.00/gridSettings.DX);
			int sourceY = (int) Math.Round(1.25/gridSettings.DX);
			int sourceZ = (int) Math.Round(1.50/gridSettings.DX);
			_sources = new[]
			           	{
			           		new Source(
			           			sourceX,
			           			sourceY,
			           			sourceZ, 440f),
			           	};
			trackBarSourceSPL.Value = (int)Toolbox.PressureTodBrefSPL0(_sources[0].PressureAmplitude);
			_microphones = new[]
			               	{
			               		new Microphone(sourceY, sourceX, sourceZ)
			               	};

			//if (_simulationSpace != null)
			//_simulationSpace.Dispose();
			try
			{
				_simulationSpace =
					new CSimulationSpace(Extent.KindE, _gpu, _sources, _microphones, gridSettings);
			}
			catch (OutOfMemoryException)
			{
				textBoxDebugOutput.Text +=
					"CUDA is out of memory -- cannot create sound volume. Try restarting the application." + Environment.NewLine;
				return;
			}
			_olaCufft.SetSimulationSpace(_simulationSpace);

			// put material from cell list into material field
			_simulationSpace.SetMaterialFrom(_materialModel);
			_simulationSpace.WallAbsorberAssociations = _wallAbsorberAssociations;

			var theSimConfig = new SimConfig {DoMur = false};

			_calc = new SimpleCalc(_simulationSpace, _gpu);
			_simulationSpace.GenerationChanged += SimSpaceGenerationChanged;

			_displaySettingsZY = new DisplaySettings(Extent);
			_displaySettingsXY = new DisplaySettings(Extent);
			_displaySettingsXZ = new DisplaySettings(Extent);
			SliceDimensionChanged(this, null);
			_displaySettingsZY.Changed += FDTDRedraw;
			_displaySettingsXY.Changed += FDTDRedraw;
			_displaySettingsXZ.Changed += FDTDRedraw;

			trackBarWeight.Value = (int) _displaySettingsXZ.WeightPressureInDecibels;

			_drawMachineZY = new FDTDDrawMachine(_simulationSpace, _displaySettingsZY, _gpu);
			_drawMachineXY = new FDTDDrawMachine(_simulationSpace, _displaySettingsXY, _gpu);
			_drawMachineXZ = new FDTDDrawMachine(_simulationSpace, _displaySettingsXZ, _gpu);

			trackBarSliceX.Maximum = _displaySettingsZY.NumberOfSlices - 1;
			trackBarSliceY.Maximum = _displaySettingsXZ.NumberOfSlices - 1;
			trackBarSliceZ.Maximum = _displaySettingsXY.NumberOfSlices - 1;

			UpdateFreeMemoryLabel();
		}

		private void UpdateFreeMemoryLabel()
		{
			if (_gpu == null) return;
			ulong freeMem = _gpu.FreeMemory;
			labelFreeMem.Text = (freeMem/1024/1024).ToString(CultureInfo.InvariantCulture) + " MB";
		}

		private void SimSpaceGenerationChanged(object sender, int generation)
		{
			labelGeneration.Text = generation.ToString(CultureInfo.InvariantCulture);
		}

		private void CheckBoxMurCheckedChanged(object sender, EventArgs e)
		{
			FDTDRedraw();
		}

		private void TrackBarWeightScroll(object sender, EventArgs e)
		{
			if (_displaySettingsXY != null)
				_displaySettingsXY.WeightBothInDecibels = trackBarWeight.Value;
			if (_displaySettingsXZ != null)
				_displaySettingsXZ.WeightBothInDecibels = trackBarWeight.Value;
			if (_displaySettingsZY != null)
				_displaySettingsZY.WeightBothInDecibels = trackBarWeight.Value;
		}

		private void TrackBarSliceZ_Scroll(object sender, EventArgs e)
		{
			if (_displaySettingsXY == null)
				return;
			_displaySettingsXY.DisplaySlice = trackBarSliceZ.Value;
		}


		private void trackBarSliceX_Scroll(object sender, EventArgs e)
		{
			if (_displaySettingsZY == null)
				return;
			_displaySettingsZY.DisplaySlice = trackBarSliceX.Value;
		}

		private void trackBarSliceY_Scroll(object sender, EventArgs e)
		{
			if (_displaySettingsXZ == null)
				return;
			_displaySettingsXZ.DisplaySlice = trackBarSliceY.Value;
		}

		/// <summary>
		/// Resets the simulation by setting generation counter to 0
		/// and clearing all sound fields
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void ButtonResetClick(object sender, EventArgs e)
		{
			_simulationSpace.Reset();
			FDTDRedraw();
			if (textBoxDebugOutput != null)
				textBoxDebugOutput.Text = string.Empty;
		}

		private void RadioButtonXCheckedChanged(object sender, EventArgs e)
		{
			SliceDimensionChanged(sender, e);
		}

		private void RadioButtonYCheckedChanged(object sender, EventArgs e)
		{
			SliceDimensionChanged(sender, e);
		}

		private void RadioButtonZCheckedChanged(object sender, EventArgs e)
		{
			SliceDimensionChanged(sender, e);
		}

		private void SliceDimensionChanged(object sender, EventArgs e)
		{
			if (_displaySettingsXZ != null)
				_displaySettingsXZ.SliceDimension = SliceDimension.Y;
			if (_displaySettingsXY != null)
				_displaySettingsXY.SliceDimension = SliceDimension.Z;
			if (_displaySettingsZY != null)
				_displaySettingsZY.SliceDimension = SliceDimension.X;
			//trackBarSliceZ.Maximum = _displaySettingsXZ.NumberOfSlices - 1;
		}

		private void TrackBarSourceSPLScroll(object sender, EventArgs e)
		{
			Source a = _simulationSpace.GetSource(0);
			a.Mutate(pressureAmplitude: Toolbox.DecibelsRefSPL0ToPressure(trackBarSourceSPL.Value));
			_simulationSpace.TheSources = new[] {a};
		}


		private void Form1_Load(object sender, EventArgs e)
		{
			_openXMLFileDialog = new OpenFileDialog();
			_openXMLFileDialog.FileName="C:\\Users\\ola\\Dropbox\\prog (svn)\\Projects\\OlaCudafy1\\OlaCudafy1\\FagerengsRom2_export.xml";
			OpenXMLFileDialog_FileOk(null, null);
			textBoxDebugOutput.GotFocus += textBoxDebugOutput_GotFocus;
			_gridAssoc.Focus();

			textBoxLargestDim.Text = 7.575.ToString(CultureInfo.InvariantCulture);
			textBoxLargestDim_KeyDown(this, new KeyEventArgs(Keys.Enter));
			buttonXML_Click(this, null);
		}

		void textBoxDebugOutput_GotFocus(object sender, EventArgs e)
		{
			_gridAssoc.Focus();
			textBoxDebugOutput.GotFocus -= textBoxDebugOutput_GotFocus;
		}

		/// <summary>
		/// Extracts the filename from the dialog just OK'ed by the user
		/// and loads a material aka room model from that file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void OpenXMLFileDialog_FileOk(object sender, CancelEventArgs e)
		{
			if (_openXMLFileDialog == null)
				return;
			//textBoxDebugOutput.Text += _openXMLFileDialog.FileName + Environment.NewLine;

			_lastFileName = _openXMLFileDialog.FileName;

			// actually load material from the file
			_materialModel = new MaterialModel(_lastFileName);
			
			// apply some stuff
			trackBarSliceX.Maximum = Extent.X;
			trackBarSliceY.Maximum = Extent.Y;
			trackBarSliceZ.Maximum = Extent.Z;
			_wallAbsorberAssociations = new WallAbsorberAssociations(_materialModel.ProduceMaterialsArray());
			_bindingSource = new BindingSource {DataSource = _wallAbsorberAssociations};
			_gridAssoc.DataSource = _bindingSource;
		}

		private BindingSource _bindingSource;
		/// <summary>
		/// An absorber row was double clicked in the absorber datagridview.
		/// </summary>
		private void absorberForm_AbsorbentDoubleClickedEvent(AbsorberDatabaseEntry entry)
		{
			foreach (DataGridViewRow selectedRow in _gridAssoc.SelectedRows)
			{
				var w = (WallAbsorberAssociations) _bindingSource.DataSource;
				w.Associate(selectedRow.Index, entry);
			}
			_gridAssoc.Refresh();
			_gridAssoc.AutoResizeColumns();
		}

		/// <summary>
		/// Shows, brings forward and gives focus to the absorber datagridview form.
		/// </summary>
		private void ShowAbsorberForm()
		{
			if (_absorberForm == null || _absorberForm.IsDisposed)
			{
				_absorberForm = new FormAbsorberDatabase();
				_absorberForm.AbsorbentDoubleClickedEvent +=
					absorberForm_AbsorbentDoubleClickedEvent;
			}
			
			_absorberForm.Show();
			_absorberForm.Focus();
		}


		/// <summary>
		/// Starting point while developing method which invokes a kernel
		/// with one thread for each microphone, picking out the microphones 
		/// reading/value at that time
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void ButtonReadMicsClick(object sender, EventArgs e)
		{
			float[,] recordings = _simulationSpace.GetRecordings();
			const int approxAvgStringLen = 12;
			var builder = new StringBuilder(_simulationSpace.RecordingsLength*approxAvgStringLen);
			const int micNumber = 0; // todo weak
			for (int gen = 0; gen < recordings.Length; gen++)
			{
				var recording =recordings[micNumber, gen]*1000f; // only one mic
				builder.Append(recording.ToString() + Environment.NewLine);
			}
			textBoxDebugOutput.Text = builder.ToString();
		}
	
		/// <summary>
		/// Creates example extent (and material model) (instead of loading material from XML)
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void ButtonCreateExtentClick(object sender, EventArgs e)
		{
			int extentSides = 5;
			var val = (int) numericUpDownExtentSides.Value;
			if (1 <= val && val <= 128)
				extentSides = val;
			var ex = new Extent(extentSides, extentSides, extentSides);
			_materialModel = new MaterialModel(ex);
		}

		/// <summary>
		/// Sets the longest length of the room model
		/// to the value specified in textBoxLargestDim (in meters)
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void buttonSetLargestDim_Click(object sender, EventArgs e)
		{
			float val;
			bool ret = float.TryParse(textBoxLargestDim.Text, out val);
			if (ret)
			{
				labelSetOK.Text = "OK";
				_simulationSpaceLargestDimension = val;
			}
			else
			{
				labelSetOK.Text = "Could not interpret. Try again";
				_simulationSpaceLargestDimension = 0F;
			}
		}


		private void ButtonFreeMemoryClick(object sender, EventArgs e)
		{
			GC.Collect();
			_freeMemoryDelayBeforeUpdateFreeMemoryLabelTimer.Start();
		}

		private void _freeMemoryDelayBeforeUpdateFreeMemoryLabelTimer_Tick(object sender, EventArgs e)
		{
			UpdateFreeMemoryLabel();
			_freeMemoryDelayBeforeUpdateFreeMemoryLabelTimer.Stop();
		}
		
		/// <summary>
		/// Opens an XML file in Westerdiep's format, with room wall positions and materials specified
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void buttonXML_Click(object sender, EventArgs e)
		{
			_openXMLFileDialog = new OpenFileDialog
			                     	{
			                     		InitialDirectory = Application.StartupPath,
			                     		CheckFileExists = true,
			                     		ReadOnlyChecked = false,
			                     		DefaultExt = "xml",
			                     		Title = Resources.Form1_Button2Click_Open_XML_file_describing_FDTD_simulation_material_walls,
			                     		FileName = _lastFileName
			                     	};
			_openXMLFileDialog.FileOk += OpenXMLFileDialog_FileOk;
			_openXMLFileDialog.ShowDialog(); // see continuing code in OpenFileDialogFileOk
		}

		private void dataGridView1_CellDoubleClick(object sender, DataGridViewCellEventArgs e)
		{
			ShowAbsorberForm();
		}

		private void dataGridView1_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyData == Keys.Enter)
			{
				ShowAbsorberForm();
				e.Handled = true;
			}
		}

		private void _gridAssoc_KeyPress(object sender, KeyPressEventArgs e)
		{
			if (Char.IsNumber(e.KeyChar))
			{
				foreach (DataGridViewRow row in _gridAssoc.Rows)
				{
					if (row.Cells["MaterialNumber"].Value.ToString()
						.StartsWith(e.KeyChar.ToString(CultureInfo.InvariantCulture),
																 true,
																 CultureInfo.InvariantCulture))
					{
						row.Selected=!row.Selected;
						return; // stop looping
					}
				}
			}
		}

		private void textBoxLargestDim_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyData == Keys.Enter)
			{
				buttonSetLargestDim.PerformClick();
				e.SuppressKeyPress = true;
			}
		}

		private void buttonToggleSimPause_Click(object sender, EventArgs e)
		{
			if (_runSimulationTimer.Enabled)
				_runSimulationTimer.Stop();
			else
				_runSimulationTimer.Start();
		}

		const float duration = 1.089f;
		
		void runSimulationTimer_Tick(object sender, EventArgs e)
		{
			while (_runSimulationTimer.Enabled)
			{
				_calc.Execute();
				if (checkBoxDoDraw.Checked)
				{
					FDTDRedraw();
					Focus();
					Refresh();
				}
				int generationsPlan = (int) Math.Ceiling(duration/_simulationSpace.DT);
				if (_simulationSpace.Generation > generationsPlan)
					_runSimulationTimer.Stop();
				break;
			}
		}

		private void zg1_Click(object sender, EventArgs e)
		{
			var p = new PointPairList();
			p.Add(0.5, 1.0);
			zg1.GraphPane.AddStick("freq", p, Color.Red);
			zg1.Refresh();
		}

		private void _gridAssoc_SelectionChanged(object sender, EventArgs e)
		{
			var a = _gridAssoc.SelectedRows;
		}


	}
}