﻿using System;
using System.Drawing;
using System.Windows.Forms;
using ZedGraph;

namespace RoomModes
{
	public partial class RoomModesForm : Form
	{
		private const double fUpper500 = 500;
		private const double fUpper5k = 5000;

		private const double cOver2 = 340/2.0;

		private const int MaxNumberOfModesToCalculate = 5000;

		// todo return oblique, tangential and axial modes in separate lists

		private double _Lx = 5.0, _Ly = 5.0, _Lz = 5.0;
		private double _fUpperLimit = 400;
		private double _maxFUpperLimit = fUpper500;

		private bool _mouseClicked;
		private int _xOffset;
		private int _yOffset;

		public RoomModesForm()
		{
			InitializeComponent();

			zg1.PointValueEvent += zg1_PointValueEvent;
			zg1.IsShowPointValues = true;
			zg1.GraphPane.Title.Text = "Room modes";
			zg1.GraphPane.XAxis.Title.Text = "Frequency (Hz)";
			zg1.GraphPane.YAxis.Title.Text = "";
			zg1.GraphPane.YAxis.Scale.Min = 0.0;
			zg1.GraphPane.YAxis.Scale.Max = 1.2;
			zg1.GraphPane.XAxis.Type = AxisType.Log;
			zg1.GraphPane.XAxis.Scale.Min = 20;
			zg1.IsEnableVPan = false;
			zg1.GraphPane.XAxis.Scale.Max = _fUpperLimit;
			zg1.GraphPane.XAxis.Scale.IsPreventLabelOverlap = true;
			zg1.GraphPane.XAxis.Scale.IsUseTenPower = false;
			zg1.GraphPane.XAxis.Scale.MajorStep = Math.Log(2, 10);
			zg1.GraphPane.XAxis.Scale.MinorStep = Math.Log(2, 10);
			zg1.GraphPane.XAxis.Scale.Mag = 2;

			zg1.ZoomEvent += zg1_ZoomEvent;
			zg1.IsEnableVZoom = false;
			UpdateModes();
			zg1.ZoomPane(zg1.GraphPane, 1.0, new PointF(), false);

			trackBar1.Value = (int) (_Lz*10.0);
		}

		private void zg1_ZoomEvent(ZedGraphControl sender, ZoomState oldState, ZoomState newState)
		{
			_fUpperLimit = zg1.GraphPane.XAxis.Scale.Max;
			if (_fUpperLimit > _maxFUpperLimit)
				_fUpperLimit = _maxFUpperLimit;
			UpdateModes();
			labelRightmostFrequency.Text = string.Format("{0:0}", zg1.GraphPane.XAxis.Scale.Max);
		}

		private double FrequencyOfMode(double Lx, double Ly, double Lz, ModeTuple mode)
		{
			return cOver2*Math.Sqrt(
				Math.Pow(mode.nx/Lx, 2.0) +
				Math.Pow(mode.ny/Ly, 2.0) +
				Math.Pow(mode.nz/Lz, 2.0));
		}

		private ModePointList[] CalculateRoomModes(double Lx, double Ly, double Lz)
		{
			var axialList = new ModePointList();
			var tangentialList = new ModePointList();
			var obliqueList = new ModePointList();
			int count = 0;
			var all = new[] {axialList, tangentialList, obliqueList};

			for (int nx = 0;; nx++)
			{
				int ny;
				for (ny = 0;; ny++)
				{
					int nz;
					for (nz = 0;; nz++)
					{
						var mode = new ModeTuple(nx, ny, nz);
						if (count > MaxNumberOfModesToCalculate)
							return all; // todo must traverse nx, ny, nz properly. bug bug bug
						double f = FrequencyOfMode(Lx, Ly, Lz, mode);
						if (f < _fUpperLimit)
						{
							if ((nx == 0 && ny == 0) ||
							    (ny == 0 && nz == 0) ||
							    (nz == 0 && nx == 0))
								axialList.Add(f, 1.0, mode);
							else if (nx == 0 || ny == 0 || nz == 0)
								tangentialList.Add(f, 1.0, mode);
							else
								obliqueList.Add(f, 1.0, mode);
							++count;
						}
						else
							break;
					}
					if (nz == 0)
						break;
				}
				if (ny == 0)
					break;
			}
			return all;
		}

		private string zg1_PointValueEvent(ZedGraphControl sender, GraphPane pane, CurveItem curve, int iPt)
		{
			return ((ModePointList) curve.Points).GetModeStringOf(iPt);
		}

		private void UpdateModes()
		{
			labelVolumeSize.Text = string.Format("{0:0.00} x {1:0.00} x {2:0.00}", _Lx, _Ly, _Lz);

			zg1.GraphPane.CurveList.Clear();

			_Lx = pictureBox1.Width/250.0*5.0;
			_Ly = pictureBox1.Height/250.0*5.0;

			ModePointList[] spikesLists = CalculateRoomModes(_Lx, _Ly, _Lz);
			ModePointList axialList = spikesLists[0];
			ModePointList tangentialList = spikesLists[1];
			ModePointList obliqueList = spikesLists[2];
			zg1.GraphPane.AddStick("Axial modes", axialList, Color.Red);
			zg1.GraphPane.AddStick("Tangential modes", tangentialList, Color.Green);
			zg1.GraphPane.AddStick("Oblique modes", obliqueList, Color.Blue);

			labelModeCount.Text = (
			                      	axialList.Count +
			                      	tangentialList.Count +
			                      	obliqueList.Count).ToString();

			zg1.Invalidate();
		}


		private void trackBar1_Scroll(object sender, EventArgs e)
		{
			_Lz = trackBar1.Value/10.0;
			UpdateModes();
		}

		private void checkBox1_CheckedChanged(object sender, EventArgs e)
		{
			if (checkBox1.Checked)
				_maxFUpperLimit = fUpper5k;
			else
				_maxFUpperLimit = fUpper500;
			zg1_ZoomEvent(null, null, null);
		}

		private void pictureBox1_MouseDown(object sender, MouseEventArgs e)
		{
			_xOffset = e.X;
			_yOffset = e.Y;
			_mouseClicked = true;
		}

		private void pictureBox1_MouseUp(object sender, MouseEventArgs e)
		{
			_mouseClicked = false;
		}

		private void pictureBox1_MouseMove_1(object sender,
		                                     MouseEventArgs e)
		{
			if (_mouseClicked)
			{
				int xDelta = e.X - _xOffset;
				_xOffset += xDelta;
				pictureBox1.Width += xDelta;
				int yDelta = e.Y - _yOffset;
				_yOffset += yDelta;
				pictureBox1.Height += yDelta;
			}

			UpdateModes();
		}
	}
}